#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import time
from collections import OrderedDict
from datetime import datetime
from typing import Optional, Set

from sqlalchemy import and_
from sqlalchemy.orm.session import Session, make_transient
from tabulate import tabulate

from airflow import models
from airflow.exceptions import (
    AirflowException,
    BackfillUnfinished,
    DagConcurrencyLimitReached,
    NoAvailablePoolSlot,
    PoolNotFound,
    TaskConcurrencyLimitReached,
)
from airflow.executors import executor_constants
from airflow.jobs.base_job import BaseJob
from airflow.models import DAG, DagPickle
from airflow.models.dagrun import DagRun
from airflow.models.taskinstance import TaskInstance, TaskInstanceKey
from airflow.ti_deps.dep_context import DepContext
from airflow.ti_deps.dependencies_deps import BACKFILL_QUEUED_DEPS
from airflow.utils import helpers, timezone
from airflow.utils.configuration import tmp_configuration_copy
from airflow.utils.session import provide_session
from airflow.utils.state import State
from airflow.utils.types import DagRunType


class BackfillJob(BaseJob):
    """
    A backfill job consists of a dag or subdag for a specific time range. It
    triggers a set of task instance runs, in the right order and lasts for
    as long as it takes for the set of task instance to be completed.
    """

    STATES_COUNT_AS_RUNNING = (State.RUNNING, State.QUEUED)

    __mapper_args__ = {'polymorphic_identity': 'BackfillJob'}

    class _DagRunTaskStatus:
        """
        Internal status of the backfill job. This class is intended to be instantiated
        only within a BackfillJob instance and will track the execution of tasks,
        e.g. running, skipped, succeeded, failed, etc. Information about the dag runs
        related to the backfill job are also being tracked in this structure,
        .e.g finished runs, etc. Any other status related information related to the
        execution of dag runs / tasks can be included in this structure since it makes
        it easier to pass it around.

        :param to_run: Tasks to run in the backfill
        :type to_run: dict[tuple[TaskInstanceKey], airflow.models.TaskInstance]
        :param running: Maps running task instance key to task instance object
        :type running: dict[tuple[TaskInstanceKey], airflow.models.TaskInstance]
        :param skipped: Tasks that have been skipped
        :type skipped: set[tuple[TaskInstanceKey]]
        :param succeeded: Tasks that have succeeded so far
        :type succeeded: set[tuple[TaskInstanceKey]]
        :param failed: Tasks that have failed
        :type failed: set[tuple[TaskInstanceKey]]
        :param not_ready: Tasks not ready for execution
        :type not_ready: set[tuple[TaskInstanceKey]]
        :param deadlocked: Deadlocked tasks
        :type deadlocked: set[airflow.models.TaskInstance]
        :param active_runs: Active dag runs at a certain point in time
        :type active_runs: list[DagRun]
        :param executed_dag_run_dates: Datetime objects for the executed dag runs
        :type executed_dag_run_dates: set[datetime.datetime]
        :param finished_runs: Number of finished runs so far
        :type finished_runs: int
        :param total_runs: Number of total dag runs able to run
        :type total_runs: int
        """

        # TODO(edgarRd): AIRFLOW-1444: Add consistency check on counts
        def __init__(  # pylint: disable=too-many-arguments
            self,
            to_run=None,
            running=None,
            skipped=None,
            succeeded=None,
            failed=None,
            not_ready=None,
            deadlocked=None,
            active_runs=None,
            executed_dag_run_dates=None,
            finished_runs=0,
            total_runs=0,
        ):
            self.to_run = to_run or OrderedDict()
            self.running = running or {}
            self.skipped = skipped or set()
            self.succeeded = succeeded or set()
            self.failed = failed or set()
            self.not_ready = not_ready or set()
            self.deadlocked = deadlocked or set()
            self.active_runs = active_runs or []
            self.executed_dag_run_dates = executed_dag_run_dates or set()
            self.finished_runs = finished_runs
            self.total_runs = total_runs

    def __init__(  # pylint: disable=too-many-arguments
        self,
        dag,
        start_date=None,
        end_date=None,
        mark_success=False,
        donot_pickle=False,
        ignore_first_depends_on_past=False,
        ignore_task_deps=False,
        pool=None,
        delay_on_limit_secs=1.0,
        verbose=False,
        conf=None,
        rerun_failed_tasks=False,
        run_backwards=False,
        *args,
        **kwargs,
    ):
        """
        :param dag: DAG object.
        :type dag: airflow.models.DAG
        :param start_date: start date for the backfill date range.
        :type start_date: datetime.datetime
        :param end_date: end date for the backfill date range.
        :type end_date: datetime.datetime
        :param mark_success: flag whether to mark the task auto success.
        :type mark_success: bool
        :param donot_pickle: whether pickle
        :type donot_pickle: bool
        :param ignore_first_depends_on_past: whether to ignore depend on past
        :type ignore_first_depends_on_past: bool
        :param ignore_task_deps: whether to ignore the task dependency
        :type ignore_task_deps: bool
        :param pool: pool to backfill
        :type pool: str
        :param delay_on_limit_secs:
        :param verbose:
        :type verbose: flag to whether display verbose message to backfill console
        :param conf: a dictionary which user could pass k-v pairs for backfill
        :type conf: dictionary
        :param rerun_failed_tasks: flag to whether to
                                   auto rerun the failed task in backfill
        :type rerun_failed_tasks: bool
        :param run_backwards: Whether to process the dates from most to least recent
        :type run_backwards bool
        :param args:
        :param kwargs:
        """
        self.dag = dag
        self.dag_id = dag.dag_id
        self.bf_start_date = start_date
        self.bf_end_date = end_date
        self.mark_success = mark_success
        self.donot_pickle = donot_pickle
        self.ignore_first_depends_on_past = ignore_first_depends_on_past
        self.ignore_task_deps = ignore_task_deps
        self.pool = pool
        self.delay_on_limit_secs = delay_on_limit_secs
        self.verbose = verbose
        self.conf = conf
        self.rerun_failed_tasks = rerun_failed_tasks
        self.run_backwards = run_backwards
        super().__init__(*args, **kwargs)

    @provide_session
    def _update_counters(self, ti_status, session=None):
        """
        Updates the counters per state of the tasks that were running. Can re-add
        to tasks to run in case required.

        :param ti_status: the internal status of the backfill job tasks
        :type ti_status: BackfillJob._DagRunTaskStatus
        """
        tis_to_be_scheduled = []
        refreshed_tis = []
        TI = TaskInstance

        filter_for_tis = TI.filter_for_tis(list(ti_status.running.values()))
        if filter_for_tis is not None:
            refreshed_tis = session.query(TI).filter(filter_for_tis).all()

        for ti in refreshed_tis:
            # Here we remake the key by subtracting 1 to match in memory information
            key = ti.key.reduced
            if ti.state == State.SUCCESS:
                ti_status.succeeded.add(key)
                self.log.debug("Task instance %s succeeded. Don't rerun.", ti)
                ti_status.running.pop(key)
                continue
            if ti.state == State.SKIPPED:
                ti_status.skipped.add(key)
                self.log.debug("Task instance %s skipped. Don't rerun.", ti)
                ti_status.running.pop(key)
                continue
            if ti.state == State.FAILED:
                self.log.error("Task instance %s failed", ti)
                ti_status.failed.add(key)
                ti_status.running.pop(key)
                continue
            # special case: if the task needs to run again put it back
            if ti.state == State.UP_FOR_RETRY:
                self.log.warning("Task instance %s is up for retry", ti)
                ti_status.running.pop(key)
                ti_status.to_run[key] = ti
            # special case: if the task needs to be rescheduled put it back
            elif ti.state == State.UP_FOR_RESCHEDULE:
                self.log.warning("Task instance %s is up for reschedule", ti)
                ti_status.running.pop(key)
                ti_status.to_run[key] = ti
            # special case: The state of the task can be set to NONE by the task itself
            # when it reaches concurrency limits. It could also happen when the state
            # is changed externally, e.g. by clearing tasks from the ui. We need to cover
            # for that as otherwise those tasks would fall outside of the scope of
            # the backfill suddenly.
            elif ti.state == State.NONE:
                self.log.warning(
                    "FIXME: task instance %s state was set to none externally or "
                    "reaching concurrency limits. Re-adding task to queue.",
                    ti,
                )
                tis_to_be_scheduled.append(ti)
                ti_status.running.pop(key)
                ti_status.to_run[key] = ti

        # Batch schedule of task instances
        if tis_to_be_scheduled:
            filter_for_tis = TI.filter_for_tis(tis_to_be_scheduled)
            session.query(TI).filter(filter_for_tis).update(
                values={TI.state: State.SCHEDULED}, synchronize_session=False
            )

    def _manage_executor_state(self, running):
        """
        Checks if the executor agrees with the state of task instances
        that are running

        :param running: dict of key, task to verify
        """
        executor = self.executor

        # TODO: query all instead of refresh from db
        for key, value in list(executor.get_event_buffer().items()):
            state, info = value
            if key not in running:
                self.log.warning("%s state %s not in running=%s", key, state, running.values())
                continue

            ti = running[key]
            ti.refresh_from_db()

            self.log.debug("Executor state: %s task %s", state, ti)

            if state in (State.FAILED, State.SUCCESS) and ti.state in self.STATES_COUNT_AS_RUNNING:
                msg = (
                    "Executor reports task instance {} finished ({}) "
                    "although the task says its {}. Was the task "
                    "killed externally? Info: {}".format(ti, state, ti.state, info)
                )
                self.log.error(msg)
                ti.handle_failure(msg)

    @provide_session
    def _get_dag_run(self, run_date: datetime, dag: DAG, session: Session = None):
        """
        Returns a dag run for the given run date, which will be matched to an existing
        dag run if available or create a new dag run otherwise. If the max_active_runs
        limit is reached, this function will return None.

        :param run_date: the execution date for the dag run
        :param dag: DAG
        :param session: the database session object
        :return: a DagRun in state RUNNING or None
        """
        # consider max_active_runs but ignore when running subdags
        respect_dag_max_active_limit = bool(dag.schedule_interval and not dag.is_subdag)

        current_active_dag_count = dag.get_num_active_runs(external_trigger=False)

        # check if we are scheduling on top of a already existing dag_run
        # we could find a "scheduled" run instead of a "backfill"
        runs = DagRun.find(dag_id=dag.dag_id, execution_date=run_date, session=session)
        run: Optional[DagRun]
        if runs:
            run = runs[0]
            if run.state == State.RUNNING:
                respect_dag_max_active_limit = False
        else:
            run = None

        # enforce max_active_runs limit for dag, special cases already
        # handled by respect_dag_max_active_limit
        if respect_dag_max_active_limit and current_active_dag_count >= dag.max_active_runs:
            return None

        run = run or dag.create_dagrun(
            execution_date=run_date,
            start_date=timezone.utcnow(),
            state=State.RUNNING,
            external_trigger=False,
            session=session,
            conf=self.conf,
            run_type=DagRunType.BACKFILL_JOB,
            creating_job_id=self.id,
        )

        # set required transient field
        run.dag = dag

        # explicitly mark as backfill and running
        run.state = State.RUNNING
        run.run_id = run.generate_run_id(DagRunType.BACKFILL_JOB, run_date)
        run.run_type = DagRunType.BACKFILL_JOB
        run.verify_integrity(session=session)
        return run

    @provide_session
    def _task_instances_for_dag_run(self, dag_run, session=None):
        """
        Returns a map of task instance key to task instance object for the tasks to
        run in the given dag run.

        :param dag_run: the dag run to get the tasks from
        :type dag_run: airflow.models.DagRun
        :param session: the database session object
        :type session: sqlalchemy.orm.session.Session
        """
        tasks_to_run = {}

        if dag_run is None:
            return tasks_to_run

        # check if we have orphaned tasks
        self.reset_state_for_orphaned_tasks(filter_by_dag_run=dag_run, session=session)

        # for some reason if we don't refresh the reference to run is lost
        dag_run.refresh_from_db()
        make_transient(dag_run)

        try:
            for ti in dag_run.get_task_instances():
                # all tasks part of the backfill are scheduled to run
                if ti.state == State.NONE:
                    ti.set_state(State.SCHEDULED, session=session)
                if ti.state != State.REMOVED:
                    tasks_to_run[ti.key] = ti
            session.commit()
        except Exception:
            session.rollback()
            raise

        return tasks_to_run

    def _log_progress(self, ti_status):
        self.log.info(
            '[backfill progress] | finished run %s of %s | tasks waiting: %s | succeeded: %s | '
            'running: %s | failed: %s | skipped: %s | deadlocked: %s | not ready: %s',
            ti_status.finished_runs,
            ti_status.total_runs,
            len(ti_status.to_run),
            len(ti_status.succeeded),
            len(ti_status.running),
            len(ti_status.failed),
            len(ti_status.skipped),
            len(ti_status.deadlocked),
            len(ti_status.not_ready),
        )

        self.log.debug("Finished dag run loop iteration. Remaining tasks %s", ti_status.to_run.values())

    @provide_session
    def _process_backfill_task_instances(  # pylint: disable=too-many-statements
        self,
        ti_status,
        executor,
        pickle_id,
        start_date=None,
        session=None,
    ):
        """
        Process a set of task instances from a set of dag runs. Special handling is done
        to account for different task instance states that could be present when running
        them in a backfill process.

        :param ti_status: the internal status of the job
        :type ti_status: BackfillJob._DagRunTaskStatus
        :param executor: the executor to run the task instances
        :type executor: BaseExecutor
        :param pickle_id: the pickle_id if dag is pickled, None otherwise
        :type pickle_id: int
        :param start_date: the start date of the backfill job
        :type start_date: datetime.datetime
        :param session: the current session object
        :type session: sqlalchemy.orm.session.Session
        :return: the list of execution_dates for the finished dag runs
        :rtype: list
        """
        executed_run_dates = []

        while (len(ti_status.to_run) > 0 or len(ti_status.running) > 0) and len(ti_status.deadlocked) == 0:
            self.log.debug("*** Clearing out not_ready list ***")
            ti_status.not_ready.clear()

            # we need to execute the tasks bottom to top
            # or leaf to root, as otherwise tasks might be
            # determined deadlocked while they are actually
            # waiting for their upstream to finish
            @provide_session
            def _per_task_process(key, ti, session=None):  # pylint: disable=too-many-return-statements
                ti.refresh_from_db(lock_for_update=True, session=session)

                task = self.dag.get_task(ti.task_id, include_subdags=True)
                ti.task = task

                ignore_depends_on_past = self.ignore_first_depends_on_past and ti.execution_date == (
                    start_date or ti.start_date
                )
                self.log.debug("Task instance to run %s state %s", ti, ti.state)

                # The task was already marked successful or skipped by a
                # different Job. Don't rerun it.
                if ti.state == State.SUCCESS:
                    ti_status.succeeded.add(key)
                    self.log.debug("Task instance %s succeeded. Don't rerun.", ti)
                    ti_status.to_run.pop(key)
                    if key in ti_status.running:
                        ti_status.running.pop(key)
                    return
                elif ti.state == State.SKIPPED:
                    ti_status.skipped.add(key)
                    self.log.debug("Task instance %s skipped. Don't rerun.", ti)
                    ti_status.to_run.pop(key)
                    if key in ti_status.running:
                        ti_status.running.pop(key)
                    return

                # guard against externally modified tasks instances or
                # in case max concurrency has been reached at task runtime
                elif ti.state == State.NONE:
                    self.log.warning(
                        "FIXME: task instance {} state was set to None " "externally. This should not happen"
                    )
                    ti.set_state(State.SCHEDULED, session=session)
                if self.rerun_failed_tasks:
                    # Rerun failed tasks or upstreamed failed tasks
                    if ti.state in (State.FAILED, State.UPSTREAM_FAILED):
                        self.log.error("Task instance %s with state %s", ti, ti.state)
                        if key in ti_status.running:
                            ti_status.running.pop(key)
                        # Reset the failed task in backfill to scheduled state
                        ti.set_state(State.SCHEDULED, session=session)
                else:
                    # Default behaviour which works for subdag.
                    if ti.state in (State.FAILED, State.UPSTREAM_FAILED):
                        self.log.error("Task instance %s with state %s", ti, ti.state)
                        ti_status.failed.add(key)
                        ti_status.to_run.pop(key)
                        if key in ti_status.running:
                            ti_status.running.pop(key)
                        return

                backfill_context = DepContext(
                    deps=BACKFILL_QUEUED_DEPS,
                    ignore_depends_on_past=ignore_depends_on_past,
                    ignore_task_deps=self.ignore_task_deps,
                    flag_upstream_failed=True,
                )

                # Is the task runnable? -- then run it
                # the dependency checker can change states of tis
                if ti.are_dependencies_met(
                    dep_context=backfill_context, session=session, verbose=self.verbose
                ):
                    if executor.has_task(ti):
                        self.log.debug(
                            "Task Instance %s already in executor " "waiting for queue to clear", ti
                        )
                    else:
                        self.log.debug('Sending %s to executor', ti)
                        # Skip scheduled state, we are executing immediately
                        ti.state = State.QUEUED
                        ti.queued_by_job_id = self.id
                        ti.queued_dttm = timezone.utcnow()
                        session.merge(ti)

                        cfg_path = None
                        if self.executor_class in (
                            executor_constants.LOCAL_EXECUTOR,
                            executor_constants.SEQUENTIAL_EXECUTOR,
                        ):
                            cfg_path = tmp_configuration_copy()

                        executor.queue_task_instance(
                            ti,
                            mark_success=self.mark_success,
                            pickle_id=pickle_id,
                            ignore_task_deps=self.ignore_task_deps,
                            ignore_depends_on_past=ignore_depends_on_past,
                            pool=self.pool,
                            cfg_path=cfg_path,
                        )
                        ti_status.running[key] = ti
                        ti_status.to_run.pop(key)
                    session.commit()
                    return

                if ti.state == State.UPSTREAM_FAILED:
                    self.log.error("Task instance %s upstream failed", ti)
                    ti_status.failed.add(key)
                    ti_status.to_run.pop(key)
                    if key in ti_status.running:
                        ti_status.running.pop(key)
                    return

                # special case
                if ti.state == State.UP_FOR_RETRY:
                    self.log.debug("Task instance %s retry period not " "expired yet", ti)
                    if key in ti_status.running:
                        ti_status.running.pop(key)
                    ti_status.to_run[key] = ti
                    return

                # special case
                if ti.state == State.UP_FOR_RESCHEDULE:
                    self.log.debug("Task instance %s reschedule period not " "expired yet", ti)
                    if key in ti_status.running:
                        ti_status.running.pop(key)
                    ti_status.to_run[key] = ti
                    return

                # all remaining tasks
                self.log.debug('Adding %s to not_ready', ti)
                ti_status.not_ready.add(key)

            try:  # pylint: disable=too-many-nested-blocks
                for task in self.dag.topological_sort(include_subdag_tasks=True):
                    for key, ti in list(ti_status.to_run.items()):
                        if task.task_id != ti.task_id:
                            continue

                        pool = session.query(models.Pool).filter(models.Pool.pool == task.pool).first()
                        if not pool:
                            raise PoolNotFound(f'Unknown pool: {task.pool}')

                        open_slots = pool.open_slots(session=session)
                        if open_slots <= 0:
                            raise NoAvailablePoolSlot(
                                "Not scheduling since there are "
                                "{} open slots in pool {}".format(open_slots, task.pool)
                            )

                        num_running_task_instances_in_dag = DAG.get_num_task_instances(
                            self.dag_id,
                            states=self.STATES_COUNT_AS_RUNNING,
                        )

                        if num_running_task_instances_in_dag >= self.dag.concurrency:
                            raise DagConcurrencyLimitReached(
                                "Not scheduling since DAG concurrency limit " "is reached."
                            )

                        if task.task_concurrency:
                            num_running_task_instances_in_task = DAG.get_num_task_instances(
                                dag_id=self.dag_id,
                                task_ids=[task.task_id],
                                states=self.STATES_COUNT_AS_RUNNING,
                            )

                            if num_running_task_instances_in_task >= task.task_concurrency:
                                raise TaskConcurrencyLimitReached(
                                    "Not scheduling since Task concurrency limit " "is reached."
                                )

                        _per_task_process(key, ti)
            except (NoAvailablePoolSlot, DagConcurrencyLimitReached, TaskConcurrencyLimitReached) as e:
                self.log.debug(e)

            # execute the tasks in the queue
            self.heartbeat()
            executor.heartbeat()

            # If the set of tasks that aren't ready ever equals the set of
            # tasks to run and there are no running tasks then the backfill
            # is deadlocked
            if (
                ti_status.not_ready
                and ti_status.not_ready == set(ti_status.to_run)
                and len(ti_status.running) == 0
            ):
                self.log.warning("Deadlock discovered for ti_status.to_run=%s", ti_status.to_run.values())
                ti_status.deadlocked.update(ti_status.to_run.values())
                ti_status.to_run.clear()

            # check executor state
            self._manage_executor_state(ti_status.running)

            # update the task counters
            self._update_counters(ti_status=ti_status)

            # update dag run state
            _dag_runs = ti_status.active_runs[:]
            for run in _dag_runs:
                run.update_state(session=session)
                if run.state in State.finished:
                    ti_status.finished_runs += 1
                    ti_status.active_runs.remove(run)
                    executed_run_dates.append(run.execution_date)

            self._log_progress(ti_status)

        # return updated status
        return executed_run_dates

    @provide_session
    def _collect_errors(self, ti_status, session=None):
        def tabulate_ti_keys_set(set_ti_keys: Set[TaskInstanceKey]) -> str:
            # Sorting by execution date first
            sorted_ti_keys = sorted(
                set_ti_keys,
                key=lambda ti_key: (ti_key.execution_date, ti_key.dag_id, ti_key.task_id, ti_key.try_number),
            )
            return tabulate(sorted_ti_keys, headers=["DAG ID", "Task ID", "Execution date", "Try number"])

        def tabulate_tis_set(set_tis: Set[TaskInstance]) -> str:
            # Sorting by execution date first
            sorted_tis = sorted(
                set_tis, key=lambda ti: (ti.execution_date, ti.dag_id, ti.task_id, ti.try_number)
            )
            tis_values = ((ti.dag_id, ti.task_id, ti.execution_date, ti.try_number) for ti in sorted_tis)
            return tabulate(tis_values, headers=["DAG ID", "Task ID", "Execution date", "Try number"])

        err = ''
        if ti_status.failed:
            err += "Some task instances failed:\n"
            err += tabulate_ti_keys_set(ti_status.failed)
        if ti_status.deadlocked:
            err += 'BackfillJob is deadlocked.'
            deadlocked_depends_on_past = any(
                t.are_dependencies_met(
                    dep_context=DepContext(ignore_depends_on_past=False),
                    session=session,
                    verbose=self.verbose,
                )
                != t.are_dependencies_met(
                    dep_context=DepContext(ignore_depends_on_past=True), session=session, verbose=self.verbose
                )
                for t in ti_status.deadlocked
            )
            if deadlocked_depends_on_past:
                err += (
                    'Some of the deadlocked tasks were unable to run because '
                    'of "depends_on_past" relationships. Try running the '
                    'backfill with the option '
                    '"ignore_first_depends_on_past=True" or passing "-I" at '
                    'the command line.'
                )
            err += '\nThese tasks have succeeded:\n'
            err += tabulate_ti_keys_set(ti_status.succeeded)
            err += '\n\nThese tasks are running:\n'
            err += tabulate_ti_keys_set(ti_status.running)
            err += '\n\nThese tasks have failed:\n'
            err += tabulate_ti_keys_set(ti_status.failed)
            err += '\n\nThese tasks are skipped:\n'
            err += tabulate_ti_keys_set(ti_status.skipped)
            err += '\n\nThese tasks are deadlocked:\n'
            err += tabulate_tis_set(ti_status.deadlocked)

        return err

    @provide_session
    def _execute_for_run_dates(self, run_dates, ti_status, executor, pickle_id, start_date, session=None):
        """
        Computes the dag runs and their respective task instances for
        the given run dates and executes the task instances.
        Returns a list of execution dates of the dag runs that were executed.

        :param run_dates: Execution dates for dag runs
        :type run_dates: list
        :param ti_status: internal BackfillJob status structure to tis track progress
        :type ti_status: BackfillJob._DagRunTaskStatus
        :param executor: the executor to use, it must be previously started
        :type executor: BaseExecutor
        :param pickle_id: numeric id of the pickled dag, None if not pickled
        :type pickle_id: int
        :param start_date: backfill start date
        :type start_date: datetime.datetime
        :param session: the current session object
        :type session: sqlalchemy.orm.session.Session
        """
        for next_run_date in run_dates:
            for dag in [self.dag] + self.dag.subdags:
                dag_run = self._get_dag_run(next_run_date, dag, session=session)
                tis_map = self._task_instances_for_dag_run(dag_run, session=session)
                if dag_run is None:
                    continue

                ti_status.active_runs.append(dag_run)
                ti_status.to_run.update(tis_map or {})

        processed_dag_run_dates = self._process_backfill_task_instances(
            ti_status=ti_status,
            executor=executor,
            pickle_id=pickle_id,
            start_date=start_date,
            session=session,
        )

        ti_status.executed_dag_run_dates.update(processed_dag_run_dates)

    @provide_session
    def _set_unfinished_dag_runs_to_failed(self, dag_runs, session=None):
        """
        Go through the dag_runs and update the state based on the task_instance state.
        Then set DAG runs that are not finished to failed.

        :param dag_runs: DAG runs
        :param session: session
        :return: None
        """
        for dag_run in dag_runs:
            dag_run.update_state()
            if dag_run.state not in State.finished:
                dag_run.set_state(State.FAILED)
            session.merge(dag_run)

    @provide_session
    def _execute(self, session=None):
        """
        Initializes all components required to run a dag for a specified date range and
        calls helper method to execute the tasks.
        """
        ti_status = BackfillJob._DagRunTaskStatus()

        start_date = self.bf_start_date

        # Get intervals between the start/end dates, which will turn into dag runs
        run_dates = self.dag.get_run_dates(start_date=start_date, end_date=self.bf_end_date)
        if self.run_backwards:
            tasks_that_depend_on_past = [t.task_id for t in self.dag.task_dict.values() if t.depends_on_past]
            if tasks_that_depend_on_past:
                raise AirflowException(
                    'You cannot backfill backwards because one or more tasks depend_on_past: {}'.format(
                        ",".join(tasks_that_depend_on_past)
                    )
                )
            run_dates = run_dates[::-1]

        if len(run_dates) == 0:
            self.log.info("No run dates were found for the given dates and dag interval.")
            return

        # picklin'
        pickle_id = None

        if not self.donot_pickle and self.executor_class not in (
            executor_constants.LOCAL_EXECUTOR,
            executor_constants.SEQUENTIAL_EXECUTOR,
            executor_constants.DASK_EXECUTOR,
        ):
            pickle = DagPickle(self.dag)
            session.add(pickle)
            session.commit()
            pickle_id = pickle.id

        executor = self.executor
        executor.start()

        ti_status.total_runs = len(run_dates)  # total dag runs in backfill

        try:  # pylint: disable=too-many-nested-blocks
            remaining_dates = ti_status.total_runs
            while remaining_dates > 0:
                dates_to_process = [
                    run_date for run_date in run_dates if run_date not in ti_status.executed_dag_run_dates
                ]

                self._execute_for_run_dates(
                    run_dates=dates_to_process,
                    ti_status=ti_status,
                    executor=executor,
                    pickle_id=pickle_id,
                    start_date=start_date,
                    session=session,
                )

                remaining_dates = ti_status.total_runs - len(ti_status.executed_dag_run_dates)
                err = self._collect_errors(ti_status=ti_status, session=session)
                if err:
                    raise BackfillUnfinished(err, ti_status)

                if remaining_dates > 0:
                    self.log.info(
                        "max_active_runs limit for dag %s has been reached "
                        " - waiting for other dag runs to finish",
                        self.dag_id,
                    )
                    time.sleep(self.delay_on_limit_secs)
        except (KeyboardInterrupt, SystemExit):
            self.log.warning("Backfill terminated by user.")

            # TODO: we will need to terminate running task instances and set the
            # state to failed.
            self._set_unfinished_dag_runs_to_failed(ti_status.active_runs)
        finally:
            session.commit()
            executor.end()

        self.log.info("Backfill done. Exiting.")

    @provide_session
    def reset_state_for_orphaned_tasks(self, filter_by_dag_run=None, session=None):
        """
        This function checks if there are any tasks in the dagrun (or all) that
        have a schedule or queued states but are not known by the executor. If
        it finds those it will reset the state to None so they will get picked
        up again.  The batch option is for performance reasons as the queries
        are made in sequence.

        :param filter_by_dag_run: the dag_run we want to process, None if all
        :type filter_by_dag_run: airflow.models.DagRun
        :return: the number of TIs reset
        :rtype: int
        """
        queued_tis = self.executor.queued_tasks
        # also consider running as the state might not have changed in the db yet
        running_tis = self.executor.running

        resettable_states = [State.SCHEDULED, State.QUEUED]
        if filter_by_dag_run is None:
            resettable_tis = (
                session.query(TaskInstance)
                .join(
                    DagRun,
                    and_(
                        TaskInstance.dag_id == DagRun.dag_id,
                        TaskInstance.execution_date == DagRun.execution_date,
                    ),
                )
                .filter(
                    # pylint: disable=comparison-with-callable
                    DagRun.state == State.RUNNING,
                    DagRun.run_type != DagRunType.BACKFILL_JOB,
                    TaskInstance.state.in_(resettable_states),
                )
            ).all()
        else:
            resettable_tis = filter_by_dag_run.get_task_instances(state=resettable_states, session=session)
        tis_to_reset = []
        # Can't use an update here since it doesn't support joins
        for ti in resettable_tis:
            if ti.key not in queued_tis and ti.key not in running_tis:
                tis_to_reset.append(ti)

        if not tis_to_reset:
            return 0

        def query(result, items):
            if not items:
                return result

            filter_for_tis = TaskInstance.filter_for_tis(items)
            reset_tis = (
                session.query(TaskInstance)
                .filter(filter_for_tis, TaskInstance.state.in_(resettable_states))
                .with_for_update()
                .all()
            )

            for ti in reset_tis:
                ti.state = State.NONE
                session.merge(ti)

            return result + reset_tis

        reset_tis = helpers.reduce_in_chunks(query, tis_to_reset, [], self.max_tis_per_query)

        task_instance_str = '\n\t'.join([repr(x) for x in reset_tis])
        session.commit()

        self.log.info("Reset the following %s TaskInstances:\n\t%s", len(reset_tis), task_instance_str)
        return len(reset_tis)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from ai_flow.project.project_config import ProjectConfig
from ai_flow.context.job_context import set_current_job_name
from ai_flow.context.project_context import init_project_config
from ai_flow.context.workflow_config_loader import init_workflow_config
from ai_flow.runtime.job_runtime_env import JobRuntimeEnv


def init_job_runtime_context(job_runtime_env: JobRuntimeEnv):
    """
    When running the ai flow job,
    you need to call this function to initialize the project, workflow and job environment information.
    This function has three functions:
    1. Init project configuration
    2. Init workflow configuration.
    3. Set current job's name.
    """
    init_project_config(job_runtime_env.project_config_file)
    init_workflow_config(job_runtime_env.workflow_config_file)
    set_current_job_name(job_runtime_env.job_name)



# 
# Licensed to the Apache Software Foundation (ASF) under one 
# or more contributor license agreements.  See the NOTICE file 
# distributed with this work for additional information 
# regarding copyright ownership.  The ASF licenses this file 
# to you under the Apache License, Version 2.0 (the 
# "License"); you may not use this file except in compliance 
# with the License.  You may obtain a copy of the License at 
# 
#   http://www.apache.org/licenses/LICENSE-2.0 
# 
# Unless required by applicable law or agreed to in writing, 
# software distributed under the License is distributed on an 
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY 
# KIND, either express or implied.  See the License for the 
# specific language governing permissions and limitations 
# under the License. 
# 

# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import scheduling_service_pb2 as scheduling__service__pb2


class SchedulingServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.submitWorkflow = channel.unary_unary(
                '/ai_flow.SchedulingService/submitWorkflow',
                request_serializer=scheduling__service__pb2.ScheduleWorkflowRequest.SerializeToString,
                response_deserializer=scheduling__service__pb2.WorkflowInfoResponse.FromString,
                )
        self.deleteWorkflow = channel.unary_unary(
                '/ai_flow.SchedulingService/deleteWorkflow',
                request_serializer=scheduling__service__pb2.ScheduleWorkflowRequest.SerializeToString,
                response_deserializer=scheduling__service__pb2.WorkflowInfoResponse.FromString,
                )
        self.pauseWorkflowScheduling = channel.unary_unary(
                '/ai_flow.SchedulingService/pauseWorkflowScheduling',
                request_serializer=scheduling__service__pb2.ScheduleWorkflowRequest.SerializeToString,
                response_deserializer=scheduling__service__pb2.WorkflowInfoResponse.FromString,
                )
        self.resumeWorkflowScheduling = channel.unary_unary(
                '/ai_flow.SchedulingService/resumeWorkflowScheduling',
                request_serializer=scheduling__service__pb2.ScheduleWorkflowRequest.SerializeToString,
                response_deserializer=scheduling__service__pb2.WorkflowInfoResponse.FromString,
                )
        self.getWorkflow = channel.unary_unary(
                '/ai_flow.SchedulingService/getWorkflow',
                request_serializer=scheduling__service__pb2.ScheduleWorkflowRequest.SerializeToString,
                response_deserializer=scheduling__service__pb2.WorkflowInfoResponse.FromString,
                )
        self.listWorkflows = channel.unary_unary(
                '/ai_flow.SchedulingService/listWorkflows',
                request_serializer=scheduling__service__pb2.ScheduleWorkflowRequest.SerializeToString,
                response_deserializer=scheduling__service__pb2.ListWorkflowInfoResponse.FromString,
                )
        self.startNewWorkflowExecution = channel.unary_unary(
                '/ai_flow.SchedulingService/startNewWorkflowExecution',
                request_serializer=scheduling__service__pb2.WorkflowExecutionRequest.SerializeToString,
                response_deserializer=scheduling__service__pb2.WorkflowExecutionResponse.FromString,
                )
        self.killAllWorkflowExecutions = channel.unary_unary(
                '/ai_flow.SchedulingService/killAllWorkflowExecutions',
                request_serializer=scheduling__service__pb2.WorkflowExecutionRequest.SerializeToString,
                response_deserializer=scheduling__service__pb2.ListWorkflowExecutionResponse.FromString,
                )
        self.killWorkflowExecution = channel.unary_unary(
                '/ai_flow.SchedulingService/killWorkflowExecution',
                request_serializer=scheduling__service__pb2.WorkflowExecutionRequest.SerializeToString,
                response_deserializer=scheduling__service__pb2.WorkflowExecutionResponse.FromString,
                )
        self.getWorkflowExecution = channel.unary_unary(
                '/ai_flow.SchedulingService/getWorkflowExecution',
                request_serializer=scheduling__service__pb2.WorkflowExecutionRequest.SerializeToString,
                response_deserializer=scheduling__service__pb2.WorkflowExecutionResponse.FromString,
                )
        self.listWorkflowExecutions = channel.unary_unary(
                '/ai_flow.SchedulingService/listWorkflowExecutions',
                request_serializer=scheduling__service__pb2.WorkflowExecutionRequest.SerializeToString,
                response_deserializer=scheduling__service__pb2.ListWorkflowExecutionResponse.FromString,
                )
        self.startJob = channel.unary_unary(
                '/ai_flow.SchedulingService/startJob',
                request_serializer=scheduling__service__pb2.ScheduleJobRequest.SerializeToString,
                response_deserializer=scheduling__service__pb2.JobInfoResponse.FromString,
                )
        self.stopJob = channel.unary_unary(
                '/ai_flow.SchedulingService/stopJob',
                request_serializer=scheduling__service__pb2.ScheduleJobRequest.SerializeToString,
                response_deserializer=scheduling__service__pb2.JobInfoResponse.FromString,
                )
        self.restartJob = channel.unary_unary(
                '/ai_flow.SchedulingService/restartJob',
                request_serializer=scheduling__service__pb2.ScheduleJobRequest.SerializeToString,
                response_deserializer=scheduling__service__pb2.JobInfoResponse.FromString,
                )
        self.getJob = channel.unary_unary(
                '/ai_flow.SchedulingService/getJob',
                request_serializer=scheduling__service__pb2.ScheduleJobRequest.SerializeToString,
                response_deserializer=scheduling__service__pb2.JobInfoResponse.FromString,
                )
        self.listJobs = channel.unary_unary(
                '/ai_flow.SchedulingService/listJobs',
                request_serializer=scheduling__service__pb2.ScheduleJobRequest.SerializeToString,
                response_deserializer=scheduling__service__pb2.ListJobInfoResponse.FromString,
                )


class SchedulingServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def submitWorkflow(self, request, context):
        """workflow api
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def deleteWorkflow(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def pauseWorkflowScheduling(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def resumeWorkflowScheduling(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getWorkflow(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def listWorkflows(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def startNewWorkflowExecution(self, request, context):
        """workflow execution api
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def killAllWorkflowExecutions(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def killWorkflowExecution(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getWorkflowExecution(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def listWorkflowExecutions(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def startJob(self, request, context):
        """job api
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def stopJob(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def restartJob(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getJob(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def listJobs(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_SchedulingServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'submitWorkflow': grpc.unary_unary_rpc_method_handler(
                    servicer.submitWorkflow,
                    request_deserializer=scheduling__service__pb2.ScheduleWorkflowRequest.FromString,
                    response_serializer=scheduling__service__pb2.WorkflowInfoResponse.SerializeToString,
            ),
            'deleteWorkflow': grpc.unary_unary_rpc_method_handler(
                    servicer.deleteWorkflow,
                    request_deserializer=scheduling__service__pb2.ScheduleWorkflowRequest.FromString,
                    response_serializer=scheduling__service__pb2.WorkflowInfoResponse.SerializeToString,
            ),
            'pauseWorkflowScheduling': grpc.unary_unary_rpc_method_handler(
                    servicer.pauseWorkflowScheduling,
                    request_deserializer=scheduling__service__pb2.ScheduleWorkflowRequest.FromString,
                    response_serializer=scheduling__service__pb2.WorkflowInfoResponse.SerializeToString,
            ),
            'resumeWorkflowScheduling': grpc.unary_unary_rpc_method_handler(
                    servicer.resumeWorkflowScheduling,
                    request_deserializer=scheduling__service__pb2.ScheduleWorkflowRequest.FromString,
                    response_serializer=scheduling__service__pb2.WorkflowInfoResponse.SerializeToString,
            ),
            'getWorkflow': grpc.unary_unary_rpc_method_handler(
                    servicer.getWorkflow,
                    request_deserializer=scheduling__service__pb2.ScheduleWorkflowRequest.FromString,
                    response_serializer=scheduling__service__pb2.WorkflowInfoResponse.SerializeToString,
            ),
            'listWorkflows': grpc.unary_unary_rpc_method_handler(
                    servicer.listWorkflows,
                    request_deserializer=scheduling__service__pb2.ScheduleWorkflowRequest.FromString,
                    response_serializer=scheduling__service__pb2.ListWorkflowInfoResponse.SerializeToString,
            ),
            'startNewWorkflowExecution': grpc.unary_unary_rpc_method_handler(
                    servicer.startNewWorkflowExecution,
                    request_deserializer=scheduling__service__pb2.WorkflowExecutionRequest.FromString,
                    response_serializer=scheduling__service__pb2.WorkflowExecutionResponse.SerializeToString,
            ),
            'killAllWorkflowExecutions': grpc.unary_unary_rpc_method_handler(
                    servicer.killAllWorkflowExecutions,
                    request_deserializer=scheduling__service__pb2.WorkflowExecutionRequest.FromString,
                    response_serializer=scheduling__service__pb2.ListWorkflowExecutionResponse.SerializeToString,
            ),
            'killWorkflowExecution': grpc.unary_unary_rpc_method_handler(
                    servicer.killWorkflowExecution,
                    request_deserializer=scheduling__service__pb2.WorkflowExecutionRequest.FromString,
                    response_serializer=scheduling__service__pb2.WorkflowExecutionResponse.SerializeToString,
            ),
            'getWorkflowExecution': grpc.unary_unary_rpc_method_handler(
                    servicer.getWorkflowExecution,
                    request_deserializer=scheduling__service__pb2.WorkflowExecutionRequest.FromString,
                    response_serializer=scheduling__service__pb2.WorkflowExecutionResponse.SerializeToString,
            ),
            'listWorkflowExecutions': grpc.unary_unary_rpc_method_handler(
                    servicer.listWorkflowExecutions,
                    request_deserializer=scheduling__service__pb2.WorkflowExecutionRequest.FromString,
                    response_serializer=scheduling__service__pb2.ListWorkflowExecutionResponse.SerializeToString,
            ),
            'startJob': grpc.unary_unary_rpc_method_handler(
                    servicer.startJob,
                    request_deserializer=scheduling__service__pb2.ScheduleJobRequest.FromString,
                    response_serializer=scheduling__service__pb2.JobInfoResponse.SerializeToString,
            ),
            'stopJob': grpc.unary_unary_rpc_method_handler(
                    servicer.stopJob,
                    request_deserializer=scheduling__service__pb2.ScheduleJobRequest.FromString,
                    response_serializer=scheduling__service__pb2.JobInfoResponse.SerializeToString,
            ),
            'restartJob': grpc.unary_unary_rpc_method_handler(
                    servicer.restartJob,
                    request_deserializer=scheduling__service__pb2.ScheduleJobRequest.FromString,
                    response_serializer=scheduling__service__pb2.JobInfoResponse.SerializeToString,
            ),
            'getJob': grpc.unary_unary_rpc_method_handler(
                    servicer.getJob,
                    request_deserializer=scheduling__service__pb2.ScheduleJobRequest.FromString,
                    response_serializer=scheduling__service__pb2.JobInfoResponse.SerializeToString,
            ),
            'listJobs': grpc.unary_unary_rpc_method_handler(
                    servicer.listJobs,
                    request_deserializer=scheduling__service__pb2.ScheduleJobRequest.FromString,
                    response_serializer=scheduling__service__pb2.ListJobInfoResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'ai_flow.SchedulingService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class SchedulingService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def submitWorkflow(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.SchedulingService/submitWorkflow',
            scheduling__service__pb2.ScheduleWorkflowRequest.SerializeToString,
            scheduling__service__pb2.WorkflowInfoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def deleteWorkflow(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.SchedulingService/deleteWorkflow',
            scheduling__service__pb2.ScheduleWorkflowRequest.SerializeToString,
            scheduling__service__pb2.WorkflowInfoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def pauseWorkflowScheduling(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.SchedulingService/pauseWorkflowScheduling',
            scheduling__service__pb2.ScheduleWorkflowRequest.SerializeToString,
            scheduling__service__pb2.WorkflowInfoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def resumeWorkflowScheduling(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.SchedulingService/resumeWorkflowScheduling',
            scheduling__service__pb2.ScheduleWorkflowRequest.SerializeToString,
            scheduling__service__pb2.WorkflowInfoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getWorkflow(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.SchedulingService/getWorkflow',
            scheduling__service__pb2.ScheduleWorkflowRequest.SerializeToString,
            scheduling__service__pb2.WorkflowInfoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def listWorkflows(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.SchedulingService/listWorkflows',
            scheduling__service__pb2.ScheduleWorkflowRequest.SerializeToString,
            scheduling__service__pb2.ListWorkflowInfoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def startNewWorkflowExecution(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.SchedulingService/startNewWorkflowExecution',
            scheduling__service__pb2.WorkflowExecutionRequest.SerializeToString,
            scheduling__service__pb2.WorkflowExecutionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def killAllWorkflowExecutions(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.SchedulingService/killAllWorkflowExecutions',
            scheduling__service__pb2.WorkflowExecutionRequest.SerializeToString,
            scheduling__service__pb2.ListWorkflowExecutionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def killWorkflowExecution(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.SchedulingService/killWorkflowExecution',
            scheduling__service__pb2.WorkflowExecutionRequest.SerializeToString,
            scheduling__service__pb2.WorkflowExecutionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getWorkflowExecution(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.SchedulingService/getWorkflowExecution',
            scheduling__service__pb2.WorkflowExecutionRequest.SerializeToString,
            scheduling__service__pb2.WorkflowExecutionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def listWorkflowExecutions(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.SchedulingService/listWorkflowExecutions',
            scheduling__service__pb2.WorkflowExecutionRequest.SerializeToString,
            scheduling__service__pb2.ListWorkflowExecutionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def startJob(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.SchedulingService/startJob',
            scheduling__service__pb2.ScheduleJobRequest.SerializeToString,
            scheduling__service__pb2.JobInfoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def stopJob(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.SchedulingService/stopJob',
            scheduling__service__pb2.ScheduleJobRequest.SerializeToString,
            scheduling__service__pb2.JobInfoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def restartJob(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.SchedulingService/restartJob',
            scheduling__service__pb2.ScheduleJobRequest.SerializeToString,
            scheduling__service__pb2.JobInfoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getJob(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.SchedulingService/getJob',
            scheduling__service__pb2.ScheduleJobRequest.SerializeToString,
            scheduling__service__pb2.JobInfoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def listJobs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.SchedulingService/listJobs',
            scheduling__service__pb2.ScheduleJobRequest.SerializeToString,
            scheduling__service__pb2.ListJobInfoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import logging
import threading
import time
from functools import wraps
from random import shuffle

import grpc
from notification_service.base_notification import BaseEvent
from notification_service.client import NotificationClient
from ai_flow.context.project_context import current_project_config
from ai_flow.endpoint.client.metadata_client import MetadataClient
from ai_flow.endpoint.client.metric_client import MetricClient
from ai_flow.endpoint.client.model_center_client import ModelCenterClient
from ai_flow.endpoint.client.scheduler_client import SchedulerClient
from ai_flow.endpoint.server.high_availability import proto_to_member, sleep_and_detecting_running
from ai_flow.project.project_config import ProjectConfig
from ai_flow.protobuf.high_availability_pb2 import ListMembersRequest, ReturnStatus
from ai_flow.protobuf.high_availability_pb2_grpc import HighAvailabilityManagerStub
from ai_flow.protobuf.metadata_service_pb2_grpc import MetadataServiceStub
from ai_flow.protobuf.metric_service_pb2_grpc import MetricServiceStub
from ai_flow.protobuf.model_center_service_pb2_grpc import ModelCenterServiceStub
from ai_flow.protobuf.scheduling_service_pb2_grpc import SchedulingServiceStub

if not hasattr(time, 'time_ns'):
    time.time_ns = lambda: int(time.time() * 1e9)

AI_FLOW_TYPE = "AI_FLOW"

_SERVER_URI = 'localhost:50051'

_default_ai_flow_client = None
_default_server_uri = 'localhost:50051'

_default_airflow_operation_client = None


def get_ai_flow_client():
    """ Get AI flow Client. """

    global _default_ai_flow_client, _default_server_uri
    if _default_ai_flow_client is None:
        current_uri = current_project_config().get_server_uri()
        if current_uri is None:
            return None
        else:
            _default_server_uri = current_uri
            _default_ai_flow_client \
                = AIFlowClient(server_uri=_default_server_uri,
                               notification_service_uri=current_project_config().get_notification_service_uri())
            return _default_ai_flow_client
    else:
        current_uri = current_project_config().get_server_uri()
        if current_uri != _default_server_uri:
            _default_server_uri = current_uri
            _default_ai_flow_client \
                = AIFlowClient(server_uri=_default_server_uri,
                               notification_service_uri=current_project_config().get_notification_service_uri())

        return _default_ai_flow_client


class AIFlowClient(MetadataClient, ModelCenterClient, NotificationClient, MetricClient, SchedulerClient):
    """
    Client of an AIFlow Server that manages metadata store, model center and notification service.
    """

    def __init__(self,
                 server_uri=_SERVER_URI,
                 notification_service_uri=None,
                 project_config: ProjectConfig = None):
        MetadataClient.__init__(self, server_uri)
        ModelCenterClient.__init__(self, server_uri)
        MetricClient.__init__(self, server_uri)
        SchedulerClient.__init__(self, server_uri)
        self.enable_ha = False
        self.list_member_interval_ms = 5000
        self.retry_interval_ms = 1000
        self.retry_timeout_ms = 10000
        if project_config is not None:
            if server_uri is None:
                server_uri = project_config.get_server_uri()
            if notification_service_uri is None:
                notification_service_uri = project_config.get_notification_service_uri()
            self.enable_ha = project_config.get_enable_ha()
            self.list_member_interval_ms = project_config.get_list_member_interval_ms()
            self.retry_interval_ms = project_config.get_retry_interval_ms()
            self.retry_timeout_ms = project_config.get_retry_timeout_ms()
        if notification_service_uri is None:
            NotificationClient.__init__(
                self,
                server_uri,
                enable_ha=self.enable_ha,
                list_member_interval_ms=self.list_member_interval_ms,
                retry_interval_ms=self.retry_interval_ms,
                retry_timeout_ms=self.retry_timeout_ms)
        else:
            NotificationClient.__init__(
                self,
                notification_service_uri,
                enable_ha=self.enable_ha,
                list_member_interval_ms=self.list_member_interval_ms,
                retry_interval_ms=self.retry_interval_ms,
                retry_timeout_ms=self.retry_timeout_ms)
        if self.enable_ha:
            server_uris = server_uri.split(",")
            self.living_aiflow_members = []
            self.current_aiflow_uri = None
            last_error = None
            for server_uri in server_uris:
                channel = grpc.insecure_channel(server_uri)
                high_availability_stub = HighAvailabilityManagerStub(channel)
                try:
                    request = ListMembersRequest(timeout_seconds=0)
                    response = high_availability_stub.listMembers(request)
                    if response.return_code == ReturnStatus.CALL_SUCCESS:
                        self.living_aiflow_members = [proto_to_member(proto).server_uri
                                                      for proto in response.members]
                    else:
                        raise Exception(response.return_msg)
                    self.current_aiflow_uri = server_uri
                    self.high_availability_stub = high_availability_stub
                    break
                except grpc.RpcError as e:
                    last_error = e
            if self.current_aiflow_uri is None:
                raise Exception("No available aiflow server uri!") from last_error
            self.aiflow_ha_change_lock = threading.Lock()
            self.aiflow_ha_running = True
            self._replace_aiflow_stubs(self.current_aiflow_uri)
            self.list_aiflow_member_thread = threading.Thread(target=self._list_aiflow_members, daemon=True)
            self.list_aiflow_member_thread.start()

    def publish_event(self, key: str, value: str, event_type: str = AI_FLOW_TYPE) -> BaseEvent:
        return self.send_event(BaseEvent(key, value, event_type))

    def _list_aiflow_members(self):
        while self.aiflow_ha_running:
            # refresh the living members
            request = ListMembersRequest(timeout_seconds=int(self.list_member_interval_ms / 1000))
            response = self.high_availability_stub.listMembers(request)
            if response.return_code == ReturnStatus.CALL_SUCCESS:
                with self.aiflow_ha_change_lock:
                    self.living_aiflow_members = [proto_to_member(proto).server_uri
                                                  for proto in response.members]
            else:
                logging.error("Exception thrown when updating the living members: %s" %
                              response.return_msg)

    def _aiflow_ha_wrapper(self, func, stub_name):
        @wraps(func)
        def call_with_retry(*args, **kwargs):
            current_stub = getattr(self, stub_name)
            current_func = getattr(current_stub, func.__name__).inner_func
            start_time = time.time_ns() / 1000000
            failed_members = set()
            while True:
                try:
                    return current_func(*args, **kwargs)
                except grpc.RpcError:
                    logging.error("Exception thrown when calling rpc, change the connection.",
                                  exc_info=True)
                    with self.aiflow_ha_change_lock:
                        # check the current_uri to ensure thread safety
                        if current_func.server_uri == self.current_aiflow_uri:
                            living_members = list(self.living_aiflow_members)
                            failed_members.add(self.current_aiflow_uri)
                            shuffle(living_members)
                            found_new_member = False
                            for server_uri in living_members:
                                if server_uri in failed_members:
                                    continue
                                next_uri = server_uri

                                self._replace_aiflow_stubs(next_uri)

                                current_func = getattr(getattr(self, stub_name),
                                                       current_func.__name__).inner_func
                                self.current_aiflow_uri = next_uri
                                found_new_member = True
                            if not found_new_member:
                                logging.error("No available living members currently. Sleep and retry.")
                                failed_members.clear()
                                sleep_and_detecting_running(self.retry_interval_ms,
                                                            lambda: self.aiflow_ha_running)

                # break if stopped or timeout
                if not self.aiflow_ha_running or \
                        time.time_ns() / 1000000 > start_time + self.retry_timeout_ms:
                    if not self.aiflow_ha_running:
                        raise Exception("HA has been disabled.")
                    else:
                        raise Exception("Rpc retry timeout!")

        call_with_retry.inner_func = func
        return call_with_retry

    def _wrap_aiflow_rpcs(self, stub, server_uri, stub_name):
        for method_name, method in dict(stub.__dict__).items():
            method.__name__ = method_name
            method.server_uri = server_uri
            setattr(stub, method_name, self._aiflow_ha_wrapper(method, stub_name))
        return stub

    def _replace_aiflow_stubs(self, server_uri):
        high_availability_channel = grpc.insecure_channel(server_uri)
        high_availability_stub = self._wrap_aiflow_rpcs(
            HighAvailabilityManagerStub(high_availability_channel),
            server_uri,
            "high_availability_stub")
        self.high_availability_stub = high_availability_stub

        metadata_channel = grpc.insecure_channel(server_uri)
        metadata_store_stub = self._wrap_aiflow_rpcs(
            MetadataServiceStub(metadata_channel),
            server_uri,
            "metadata_store_stub")
        self.metadata_store_stub = metadata_store_stub

        model_center_channel = grpc.insecure_channel(server_uri)
        model_center_stub = self._wrap_aiflow_rpcs(
            ModelCenterServiceStub(model_center_channel),
            server_uri,
            "model_center_stub")
        self.model_center_stub = model_center_stub

        metric_channel = grpc.insecure_channel(server_uri)
        metric_stub = self._wrap_aiflow_rpcs(
            MetricServiceStub(metric_channel),
            server_uri,
            "metric_stub")
        self.metric_stub = metric_stub

        scheduling_channel = grpc.insecure_channel(server_uri)
        scheduling_stub = self._wrap_aiflow_rpcs(
            SchedulingServiceStub(scheduling_channel),
            server_uri,
            "scheduling_stub")
        self.scheduling_stub = scheduling_stub

    def disable_high_availability(self):
        if hasattr(self, "aiflow_ha_running"):
            self.aiflow_ha_running = False
        NotificationClient.disable_high_availability(self)
        if hasattr(self, "aiflow_ha_running"):
            self.list_aiflow_member_thread.join()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module is deprecated. Please use `airflow.providers.google.cloud.operators.cloud_sql`."""

import warnings

from airflow.providers.google.cloud.operators.cloud_sql import (
    CloudSQLBaseOperator,
    CloudSQLCreateInstanceDatabaseOperator,
    CloudSQLCreateInstanceOperator,
    CloudSQLDeleteInstanceDatabaseOperator,
    CloudSQLDeleteInstanceOperator,
    CloudSQLExecuteQueryOperator,
    CloudSQLExportInstanceOperator,
    CloudSQLImportInstanceOperator,
    CloudSQLInstancePatchOperator,
    CloudSQLPatchInstanceDatabaseOperator,
)

warnings.warn(
    "This module is deprecated. Please use `airflow.providers.google.cloud.operators.cloud_sql`",
    DeprecationWarning,
    stacklevel=2,
)


class CloudSqlBaseOperator(CloudSQLBaseOperator):
    """
    This class is deprecated. Please use
    `airflow.providers.google.cloud.operators.sql.CloudSQLBaseOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(self.__doc__, DeprecationWarning, stacklevel=3)
        super().__init__(*args, **kwargs)


class CloudSqlInstanceCreateOperator(CloudSQLCreateInstanceOperator):
    """
    This class is deprecated. Please use `airflow.providers.google.cloud.operators.sql
    .CloudSQLCreateInstanceOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(self.__doc__, DeprecationWarning, stacklevel=3)
        super().__init__(*args, **kwargs)


class CloudSqlInstanceDatabaseCreateOperator(CloudSQLCreateInstanceDatabaseOperator):
    """
    This class is deprecated. Please use `airflow.providers.google.cloud.operators.sql
    .CloudSQLCreateInstanceDatabaseOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(self.__doc__, DeprecationWarning, stacklevel=3)
        super().__init__(*args, **kwargs)


class CloudSqlInstanceDatabaseDeleteOperator(CloudSQLDeleteInstanceDatabaseOperator):
    """
    This class is deprecated. Please use `airflow.providers.google.cloud.operators.sql
    .CloudSQLDeleteInstanceDatabaseOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(self.__doc__, DeprecationWarning, stacklevel=3)
        super().__init__(*args, **kwargs)


class CloudSqlInstanceDatabasePatchOperator(CloudSQLPatchInstanceDatabaseOperator):
    """
    This class is deprecated. Please use `airflow.providers.google.cloud.operators.sql
    .CloudSQLPatchInstanceDatabaseOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(self.__doc__, DeprecationWarning, stacklevel=3)
        super().__init__(*args, **kwargs)


class CloudSqlInstanceDeleteOperator(CloudSQLDeleteInstanceOperator):
    """
    This class is deprecated. Please use `airflow.providers.google.cloud.operators.sql
    .CloudSQLDeleteInstanceOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(self.__doc__, DeprecationWarning, stacklevel=3)
        super().__init__(*args, **kwargs)


class CloudSqlInstanceExportOperator(CloudSQLExportInstanceOperator):
    """
    This class is deprecated. Please use `airflow.providers.google.cloud.operators.sql
    .CloudSQLExportInstanceOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(self.__doc__, DeprecationWarning, stacklevel=3)
        super().__init__(*args, **kwargs)


class CloudSqlInstanceImportOperator(CloudSQLImportInstanceOperator):
    """
    This class is deprecated. Please use `airflow.providers.google.cloud.operators.sql
    .CloudSQLImportInstanceOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(self.__doc__, DeprecationWarning, stacklevel=3)
        super().__init__(*args, **kwargs)


class CloudSqlInstancePatchOperator(CloudSQLInstancePatchOperator):
    """
    This class is deprecated. Please use `airflow.providers.google.cloud.operators
    .sql.CloudSQLInstancePatchOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(self.__doc__, DeprecationWarning, stacklevel=3)
        super().__init__(*args, **kwargs)


class CloudSqlQueryOperator(CloudSQLExecuteQueryOperator):
    """
    This class is deprecated. Please use `airflow.providers.google.cloud.operators
    .sql.CloudSQLExecuteQueryOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(self.__doc__, DeprecationWarning, stacklevel=3)
        super().__init__(*args, **kwargs)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Increase length of FAB ab_view_menu.name column

Revision ID: 03afc6b6f902
Revises: 92c57b58940d
Create Date: 2020-11-13 22:21:41.619565

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.engine.reflection import Inspector

# revision identifiers, used by Alembic.
revision = '03afc6b6f902'
down_revision = '92c57b58940d'
branch_labels = None
depends_on = None


def upgrade():
    """Apply Increase length of FAB ab_view_menu.name column"""
    conn = op.get_bind()
    inspector = Inspector.from_engine(conn)
    tables = inspector.get_table_names()

    if "ab_view_menu" in tables:
        if conn.dialect.name == "sqlite":
            op.execute("PRAGMA foreign_keys=off")
            op.execute(
                """
            CREATE TABLE IF NOT EXISTS ab_view_menu_dg_tmp
            (
                id INTEGER NOT NULL PRIMARY KEY,
                name VARCHAR(250) NOT NULL UNIQUE
            );
            """
            )
            op.execute("INSERT INTO ab_view_menu_dg_tmp(id, name) select id, name from ab_view_menu;")
            op.execute("DROP TABLE ab_view_menu")
            op.execute("ALTER TABLE ab_view_menu_dg_tmp rename to ab_view_menu;")
            op.execute("PRAGMA foreign_keys=on")
        else:
            op.alter_column(
                table_name='ab_view_menu', column_name='name', type_=sa.String(length=250), nullable=False
            )


def downgrade():
    """Unapply Increase length of FAB ab_view_menu.name column"""
    conn = op.get_bind()
    inspector = Inspector.from_engine(conn)
    tables = inspector.get_table_names()
    if "ab_view_menu" in tables:
        if conn.dialect.name == "sqlite":
            op.execute("PRAGMA foreign_keys=off")
            op.execute(
                """
                CREATE TABLE IF NOT EXISTS ab_view_menu_dg_tmp
                (
                    id INTEGER NOT NULL PRIMARY KEY,
                    name VARCHAR(100) NOT NULL UNIQUE
                );
                """
            )
            op.execute("INSERT INTO ab_view_menu_dg_tmp(id, name) select id, name from ab_view_menu;")
            op.execute("DROP TABLE ab_view_menu")
            op.execute("ALTER TABLE ab_view_menu_dg_tmp rename to ab_view_menu;")
            op.execute("PRAGMA foreign_keys=on")
        else:
            op.alter_column(
                table_name='ab_view_menu', column_name='name', type_=sa.String(length=100), nullable=False
            )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import time
import unittest

from notification_service.event_storage import MemoryEventStorage
from notification_service.master import NotificationMaster
from notification_service.service import NotificationService

from ai_flow.project.project_config import ProjectConfig
from ai_flow.client.ai_flow_client import AIFlowClient
from ai_flow.endpoint.server.server import AIFlowServer
from ai_flow.store.db.base_model import base
from ai_flow.store.sqlalchemy_store import SqlAlchemyStore

_SQLITE_DB_FILE = 'aiflow.db'
_SQLITE_DB_URI = '%s%s' % ('sqlite:///', _SQLITE_DB_FILE)


class TestHighAvailableAIFlowServer(unittest.TestCase):

    @staticmethod
    def start_aiflow_server(host, port):
        port = str(port)
        server_uri = host + ":" + port
        server = AIFlowServer(
            store_uri=_SQLITE_DB_URI, port=port, enabled_ha=True, start_scheduler_service=False,
            ha_server_uri=server_uri, notification_uri='localhost:30031', start_default_notification=False)
        server.run()
        return server

    def wait_for_new_members_detected(self, new_member_uri):
        while True:
            living_member = self.client.living_aiflow_members
            if new_member_uri in living_member:
                break
            else:
                time.sleep(1)

    def setUp(self) -> None:
        SqlAlchemyStore(_SQLITE_DB_URI)
        self.notification = NotificationMaster(service=NotificationService(storage=MemoryEventStorage()), port=30031)
        self.notification.run()
        self.server1 = AIFlowServer(
            store_uri=_SQLITE_DB_URI, port=50051, enabled_ha=True, start_scheduler_service=False,
            ha_server_uri='localhost:50051', notification_uri='localhost:30031', start_default_notification=False)
        self.server1.run()
        self.server2 = None
        self.server3 = None
        self.config = ProjectConfig()
        self.config.set_enable_ha(True)
        self.config.set_notification_service_uri('localhost:30031')
        self.client = AIFlowClient(server_uri='localhost:50052,localhost:50051', project_config=self.config)

    def tearDown(self) -> None:
        self.client.stop_listen_event()
        self.client.disable_high_availability()
        if self.server1 is not None:
            self.server1.stop()
        if self.server2 is not None:
            self.server2.stop()
        if self.server3 is not None:
            self.server3.stop()
        if self.notification is not None:
            self.notification.stop()
        store = SqlAlchemyStore(_SQLITE_DB_URI)
        base.metadata.drop_all(store.db_engine)

    def test_server_change(self) -> None:
        self.client.register_project("test_project")
        projects = self.client.list_project(10, 0)
        self.assertEqual(self.client.current_aiflow_uri, "localhost:50051")
        self.assertEqual(projects[0].name, "test_project")

        self.server2 = self.start_aiflow_server("localhost", 50052)
        self.wait_for_new_members_detected("localhost:50052")
        self.server1.stop()
        projects = self.client.list_project(10, 0)
        self.assertEqual(self.client.current_aiflow_uri, "localhost:50052")
        self.assertEqual(projects[0].name, "test_project")

        self.server3 = self.start_aiflow_server("localhost", 50053)
        self.wait_for_new_members_detected("localhost:50053")
        self.server2.stop()
        projects = self.client.list_project(10, 0)
        self.assertEqual(self.client.current_aiflow_uri, "localhost:50053")
        self.assertEqual(projects[0].name, "test_project")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import unittest
import time
from datetime import datetime

from ai_flow.util import json_utils

from ai_flow.workflow.control_edge import SchedulingRule, MeetAnyEventCondition, EventLife, ValueCondition, JobAction, \
    MeetAllEventCondition
from airflow.executors.scheduling_action import SchedulingAction
from notification_service.base_notification import BaseEvent
from ai_flow_plugins.scheduler_plugins.airflow.event_handler import AIFlowHandler, AiFlowTs


class TestAIFlowEventHandlers(unittest.TestCase):

    def test_one_config(self):
        rule = SchedulingRule(MeetAnyEventCondition().add_event(
            event_key="key_1",
            event_value="value_1",
            sender="1-job-name",
            value_condition=ValueCondition.EQUALS), action=JobAction.START)
        configs = [rule]
        config_str = json_utils.dumps(configs)
        handler = AIFlowHandler(config=config_str)
        event: BaseEvent = BaseEvent(key='key_1',
                                     value='value_1',
                                     namespace='default',
                                     sender='1-job-name',
                                     create_time=round(time.time() * 1000))
        action, ts = handler.handle_event(event, None)
        self.assertEqual(SchedulingAction.START, action)

    def test_two_events_config(self):
        condition = MeetAllEventCondition() \
            .add_event(event_key="key_1", event_value="value_1", sender="1-job-name") \
            .add_event(event_key="key_2", event_value="value_2", sender="1-job-name")
        rule = SchedulingRule(condition, action=JobAction.START)
        configs = [rule]
        config_str = json_utils.dumps(configs)
        handler = AIFlowHandler(config=config_str)
        event: BaseEvent = BaseEvent(key='key_1',
                                     value='value_1',
                                     namespace='default',
                                     sender='1-job-name',
                                     create_time=round(time.time() * 1000))
        action, ts = handler.handle_event(event, None)
        self.assertEqual(SchedulingAction.NONE, action)

        event: BaseEvent = BaseEvent(key='key_2',
                                     value='value_2',
                                     namespace='default',
                                     sender='1-job-name',
                                     create_time=round(time.time() * 1000))
        action, ts = handler.handle_event(event, ts)
        self.assertEqual(SchedulingAction.START, action)

    def test_two_rules(self):
        condition = MeetAnyEventCondition() \
            .add_event(event_key="key_1", event_value="value_1", sender="1-job-name")
        rule1 = SchedulingRule(condition, action=JobAction.START)
        condition = MeetAllEventCondition() \
            .add_event(event_key="key_2", event_value="value_2", sender="1-job-name")
        rule2 = SchedulingRule(condition, action=JobAction.START)
        configs = [rule1, rule2]
        config_str = json_utils.dumps(configs)
        handler = AIFlowHandler(config=config_str)
        event: BaseEvent = BaseEvent(key='key_1',
                                     value='value_1',
                                     namespace='default',
                                     sender='1-job-name',
                                     create_time=round(time.time() * 1000))
        action, ts = handler.handle_event(event, None)
        self.assertEqual(SchedulingAction.START, action)

    def test_namespace_any(self):
        rule = SchedulingRule(MeetAllEventCondition().add_event(
            event_key="key_1",
            event_value="value_1",
            namespace="*",
            sender="1-job-name"), action=JobAction.START)
        configs = [rule]
        config_str = json_utils.dumps(configs)
        handler = AIFlowHandler(config=config_str)
        event: BaseEvent = BaseEvent(key='key_1',
                                     value='value_1',
                                     namespace='aa',
                                     sender='1-job-name',
                                     create_time=int(time.time() * 1000))
        action, ts = handler.handle_event(event, None)
        self.assertEqual(SchedulingAction.START, action)

        event: BaseEvent = BaseEvent(key='key_1',
                                     value='value_1',
                                     namespace='bb',
                                     sender='1-job-name',
                                     create_time=int(time.time() * 1000 + 1))
        action, ts = handler.handle_event(event, ts)
        self.assertEqual(SchedulingAction.START, action)

        rule = SchedulingRule(MeetAllEventCondition().add_event(
            event_key="key_1",
            event_value="value_1",
            namespace="aa",
            sender="1-job-name"), action=JobAction.START)

        configs = [rule]
        config_str = json_utils.dumps(configs)
        handler = AIFlowHandler(config=config_str)
        event: BaseEvent = BaseEvent(key='key_1',
                                     value='value_1',
                                     namespace='bb',
                                     sender='1-job-name',
                                     create_time=int(time.time() * 1000))
        action, ts = handler.handle_event(event, None)
        self.assertEqual(SchedulingAction.NONE, action)

    def test_event_type_any(self):
        rule = SchedulingRule(MeetAllEventCondition().add_event(
            event_key="key_1",
            event_type="*",
            event_value="value_1",
            namespace="aa",
            sender="1-job-name"), action=JobAction.START)
        configs = [rule]
        config_str = json_utils.dumps(configs)
        handler = AIFlowHandler(config=config_str)
        event: BaseEvent = BaseEvent(key='key_1',
                                     value='value_1',
                                     event_type='aa',
                                     namespace='aa',
                                     sender='1-job-name',
                                     create_time=int(time.time() * 1000))
        action, ts = handler.handle_event(event, None)
        self.assertEqual(SchedulingAction.START, action)

        event: BaseEvent = BaseEvent(key='key_1',
                                     value='value_1',
                                     namespace='aa',
                                     event_type='bb',
                                     sender='1-job-name',
                                     create_time=int(time.time() * 1000 + 1))
        action, ts = handler.handle_event(event, ts)
        self.assertEqual(SchedulingAction.START, action)

        rule = SchedulingRule(MeetAllEventCondition().add_event(
            event_key="key_1",
            event_type="aa",
            event_value="value_1",
            namespace="aa",
            sender="1-job-name"), action=JobAction.START)
        configs = [rule]
        config_str = json_utils.dumps(configs)
        handler = AIFlowHandler(config=config_str)
        event: BaseEvent = BaseEvent(key='key_1',
                                     value='value_1',
                                     event_type='bb',
                                     namespace='aa',
                                     sender='1-job-name',
                                     create_time=int(time.time() * 1000))
        action, ts = handler.handle_event(event, None)
        self.assertEqual(SchedulingAction.NONE, action)

    def test_sender_any(self):
        rule = SchedulingRule(MeetAllEventCondition().add_event(
            event_key="key_1",
            event_value="value_1",
            namespace="aa",
            sender="*"), action=JobAction.START)
        configs = [rule]
        config_str = json_utils.dumps(configs)
        handler = AIFlowHandler(config=config_str)
        event: BaseEvent = BaseEvent(key='key_1',
                                     value='value_1',
                                     namespace='aa',
                                     sender='aa',
                                     create_time=int(time.time() * 1000))
        action, ts = handler.handle_event(event, None)
        self.assertEqual(SchedulingAction.START, action)

        event: BaseEvent = BaseEvent(key='key_1',
                                     value='value_1',
                                     namespace='aa',
                                     sender='bb',
                                     create_time=int(time.time() * 1000 + 1))
        action, ts = handler.handle_event(event, None)
        self.assertEqual(SchedulingAction.START, action)

        rule = SchedulingRule(MeetAllEventCondition().add_event(
            event_key="key_1",
            event_value="value_1",
            namespace="aa",
            sender="aa"), action=JobAction.START)
        configs = [rule]
        config_str = json_utils.dumps(configs)
        handler = AIFlowHandler(config=config_str)
        event: BaseEvent = BaseEvent(key='key_1',
                                     value='value_1',
                                     namespace='aa',
                                     sender='bb',
                                     create_time=int(time.time() * 1000))
        action, ts = handler.handle_event(event, None)
        self.assertEqual(SchedulingAction.NONE, action)

    def test_key_any(self):
        rule = SchedulingRule(MeetAllEventCondition().add_event(
            event_key="*",
            event_value="value_1",
            namespace="aa",
            sender="aa"), action=JobAction.START)
        configs = [rule]
        config_str = json_utils.dumps(configs)
        handler = AIFlowHandler(config=config_str)
        event: BaseEvent = BaseEvent(key='key_1_1',
                                     value='value_1',
                                     namespace='aa',
                                     sender='aa',
                                     create_time=int(time.time() * 1000))
        action, ts = handler.handle_event(event, None)
        self.assertEqual(SchedulingAction.START, action)

        event: BaseEvent = BaseEvent(key='key_1_2',
                                     value='value_1',
                                     namespace='aa',
                                     sender='aa',
                                     create_time=int(time.time() * 1000 + 1))
        action, ts = handler.handle_event(event, ts)
        self.assertEqual(SchedulingAction.START, action)

        rule = SchedulingRule(MeetAllEventCondition().add_event(
            event_key="aa",
            event_value="value_1",
            namespace="aa",
            sender="aa"), action=JobAction.START)
        configs = [rule]
        config_str = json_utils.dumps(configs)
        handler = AIFlowHandler(config=config_str)
        event: BaseEvent = BaseEvent(key='key_1_1',
                                     value='value_1',
                                     namespace='aa',
                                     sender='aa',
                                     create_time=int(time.time() * 1000))
        action, ts = handler.handle_event(event, None)
        self.assertEqual(SchedulingAction.NONE, action)

    def test_multiple_any_config(self):
        rule = SchedulingRule(MeetAllEventCondition().add_event(
            event_key="key_1",
            event_value="value_1",
            event_type="*",
            namespace="*",
            sender="1-job-name"), action=JobAction.START)
        configs = [rule]
        config_str = json_utils.dumps(configs)
        handler = AIFlowHandler(config=config_str)
        event: BaseEvent = BaseEvent(key='key_1',
                                     value='value_1',
                                     namespace='default',
                                     sender='1-job-name',
                                     create_time=round(time.time() * 1000))
        action, ts = handler.handle_event(event, None)
        self.assertEqual(SchedulingAction.START, action)

        event: BaseEvent = BaseEvent(key='key_1',
                                     value='value_1',
                                     namespace='default',
                                     sender='aa',
                                     create_time=round(time.time() * 1000))
        action, ts = handler.handle_event(event, ts)
        self.assertEqual(SchedulingAction.NONE, action)

        event: BaseEvent = BaseEvent(key='key_1_1',
                                     value='value_1',
                                     namespace='default',
                                     sender='1-job-name',
                                     create_time=round(time.time() * 1000))
        action, ts = handler.handle_event(event, ts)
        self.assertEqual(SchedulingAction.NONE, action)

    def test_handle_task_status_change_event(self):
        from airflow.events.scheduler_events import TaskStateChangedEvent
        from airflow.utils.state import State

        rule1 = SchedulingRule(MeetAnyEventCondition().add_event(
            event_key="dag_1.task_1",
            event_value="RUNNING",
            event_type="TASK_STATUS_CHANGED",
            namespace="test",
            sender="task_1"), action=JobAction.START)
        rule2 = SchedulingRule(MeetAnyEventCondition().add_event(
            event_key="dag_1.task_1",
            event_value="FINISHED",
            event_type="TASK_STATUS_CHANGED",
            namespace="test",
            sender="task_1"), action=JobAction.STOP)
        configs = [rule1, rule2]
        config_str = json_utils.dumps(configs)
        handler = AIFlowHandler(config=config_str)

        event = TaskStateChangedEvent("task_1", "test.dag_1", datetime.utcnow(), State.RUNNING).to_event()
        action, ts = handler.handle_event(event, None)
        self.assertEqual(SchedulingAction.START, action)

        time.sleep(0.01)
        event = TaskStateChangedEvent("task_1", "test.dag_1", datetime.utcnow(), State.KILLING).to_event()
        action, ts = handler.handle_event(event, ts)
        self.assertEqual(SchedulingAction.NONE, action)

        time.sleep(0.01)
        event = TaskStateChangedEvent("task_1", "test.dag_1", datetime.utcnow(), State.SUCCESS).to_event()
        action, ts = handler.handle_event(event, ts)
        self.assertEqual(SchedulingAction.STOP, action)

    def test_event_handler_schedule_time(self):
        condition = MeetAnyEventCondition() \
            .add_event(event_key="key_1", event_value="value_1", sender="1-job-name")
        rule1 = SchedulingRule(condition, action=JobAction.START)
        condition = MeetAllEventCondition() \
            .add_event(event_key="key_1", event_value="value_1", sender="1-job-name") \
            .add_event(event_key='key_2', event_value='value_2', sender="1-job-name")
        rule2 = SchedulingRule(condition, action=JobAction.STOP)

        configs = [rule1, rule2]
        config_str = json_utils.dumps(configs)
        handler = AIFlowHandler(config=config_str)

        event1_time = round(time.time() * 1000)
        event1: BaseEvent = BaseEvent(key='key_1',
                                      value='value_1',
                                      namespace='default',
                                      sender='1-job-name',
                                      create_time=event1_time)
        action, ts = handler.handle_event(event1, None)
        ts: AiFlowTs = ts
        self.assertEqual(event1_time, ts.rule_states[0].schedule_time)
        self.assertEqual(event1_time, ts.rule_states[0].latest_time)
        self.assertNotEqual(event1_time, ts.rule_states[1].schedule_time)
        self.assertEqual(event1_time, ts.rule_states[1].latest_time)

    def test_two_rule_trigger_at_same_time_take_action_from_first_rule(self):
        condition = MeetAnyEventCondition() \
            .add_event(event_key="key_1", event_value="value_1", sender="1-job-name")
        rule1 = SchedulingRule(condition, action=JobAction.START)
        condition = MeetAllEventCondition() \
            .add_event(event_key="key_1", event_value="value_1", sender="1-job-name") \
            .add_event(event_key='key_2', event_value='value_2', sender="1-job-name")
        rule2 = SchedulingRule(condition, action=JobAction.STOP)

        configs = [rule1, rule2]
        config_str = json_utils.dumps(configs)
        handler = AIFlowHandler(config=config_str)
        event1: BaseEvent = BaseEvent(key='key_1',
                                      value='value_1',
                                      namespace='default',
                                      sender='1-job-name',
                                      create_time=round(time.time() * 1000))
        event2: BaseEvent = BaseEvent(key='key_2',
                                      value='value_2',
                                      namespace='default',
                                      sender='1-job-name',
                                      create_time=round(time.time() * 1000))
        action, ts = handler.handle_event(event2, None)
        self.assertEqual(SchedulingAction.NONE, action)

        action, ts = handler.handle_event(event1, ts)
        self.assertEqual(SchedulingAction.START, action)

    def test_two_rules_trigger_will_clear_all_event_life(self):
        condition = MeetAnyEventCondition() \
            .add_event(event_key="key_1", event_value="value_1", sender="1-job-name")
        rule1 = SchedulingRule(condition, action=JobAction.START)
        condition = MeetAllEventCondition() \
            .add_event(event_key="key_1", event_value="value_1", sender="1-job-name") \
            .add_event(event_key='key_2', event_value='value_2', sender="1-job-name")
        rule2 = SchedulingRule(condition, action=JobAction.STOP)

        configs = [rule1, rule2]
        config_str = json_utils.dumps(configs)
        handler = AIFlowHandler(config=config_str)
        event1: BaseEvent = BaseEvent(key='key_1',
                                      value='value_1',
                                      namespace='default',
                                      sender='1-job-name',
                                      create_time=round(time.time() * 1000))
        event2: BaseEvent = BaseEvent(key='key_2',
                                      value='value_2',
                                      namespace='default',
                                      sender='1-job-name',
                                      create_time=round(time.time() * 1000))
        action, ts = handler.handle_event(event2, None)
        self.assertEqual(SchedulingAction.NONE, action)
        action, ts = handler.handle_event(event1, ts)
        self.assertEqual(SchedulingAction.START, action)
        action, ts = handler.handle_event(event2, ts)
        self.assertEqual(SchedulingAction.NONE, action)

    def test_unordered_event_life_once(self):
        condition = MeetAnyEventCondition() \
            .add_event(event_key="key_1", event_value="value_1", sender="1-job-name")
        rule1 = SchedulingRule(condition, action=JobAction.START)

        configs = [rule1]
        config_str = json_utils.dumps(configs)
        handler = AIFlowHandler(config=config_str)
        event1: BaseEvent = BaseEvent(key='key_1',
                                      value='value_1',
                                      namespace='default',
                                      sender='1-job-name',
                                      create_time=round(time.time() * 1000))
        event1_in_past: BaseEvent = BaseEvent(key='key_1',
                                              value='value_1',
                                              namespace='default',
                                              sender='1-job-name',
                                              create_time=0)
        action, ts = handler.handle_event(event1, None)
        self.assertEqual(SchedulingAction.START, action)
        action, ts = handler.handle_event(event1_in_past, ts)
        self.assertEqual(SchedulingAction.NONE, action)

    def test_unordered_event_life_repeated(self):
        condition = MeetAnyEventCondition() \
            .add_event(event_key="key_1", event_value="value_1", sender="1-job-name", life=EventLife.REPEATED)
        rule1 = SchedulingRule(condition, action=JobAction.START)

        configs = [rule1]
        config_str = json_utils.dumps(configs)
        handler = AIFlowHandler(config=config_str)
        event1: BaseEvent = BaseEvent(key='key_1',
                                      value='value_1',
                                      namespace='default',
                                      sender='1-job-name',
                                      create_time=round(time.time() * 1000))
        event1_in_past: BaseEvent = BaseEvent(key='key_1',
                                              value='value_1',
                                              namespace='default',
                                              sender='1-job-name',
                                              create_time=0)
        action, ts = handler.handle_event(event1, None)
        self.assertEqual(SchedulingAction.START, action)
        action, ts = handler.handle_event(event1_in_past, ts)
        self.assertEqual(SchedulingAction.START, action)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import unittest
import os
from typing import Text, List, Optional

from ai_flow.context.project_context import ProjectContext
from ai_flow.endpoint.server.server_runner import AIFlowServerRunner
from ai_flow.plugin_interface.scheduler_interface import Scheduler, JobExecutionInfo, WorkflowExecutionInfo, \
    WorkflowInfo
from ai_flow.workflow.workflow import Workflow


class MockScheduler(Scheduler):

    def delete_workflow(self, project_name: Text, workflow_name: Text) -> WorkflowInfo:
        pass

    def submit_workflow(self, workflow: Workflow, project_context: ProjectContext) -> WorkflowInfo:
        pass

    def pause_workflow_scheduling(self, project_name: Text, workflow_name: Text) -> WorkflowInfo:
        pass

    def resume_workflow_scheduling(self, project_name: Text, workflow_name: Text) -> WorkflowInfo:
        pass

    def start_new_workflow_execution(self, project_name: Text, workflow_name: Text) -> Optional[WorkflowExecutionInfo]:
        pass

    def stop_all_workflow_execution(self, project_name: Text, workflow_name: Text) -> List[WorkflowExecutionInfo]:
        pass

    def stop_workflow_execution(self, workflow_execution_id: Text) -> Optional[WorkflowExecutionInfo]:
        pass

    def get_workflow_execution(self, workflow_execution_id: Text) -> Optional[WorkflowExecutionInfo]:
        pass

    def list_workflow_executions(self, project_name: Text, workflow_name: Text) -> List[WorkflowExecutionInfo]:
        pass

    def start_job_execution(self, job_name: Text, workflow_execution_id: Text) -> JobExecutionInfo:
        pass

    def stop_job_execution(self, job_name: Text, workflow_execution_id: Text) -> JobExecutionInfo:
        pass

    def restart_job_execution(self, job_name: Text, workflow_execution_id: Text) -> JobExecutionInfo:
        pass

    def get_job_executions(self, job_name: Text, workflow_execution_id: Text) -> List[JobExecutionInfo]:
        pass

    def list_job_executions(self, workflow_execution_id: Text) -> List[JobExecutionInfo]:
        pass


class TestServerRunner(unittest.TestCase):

    def test_server_start_stop(self):
        server_runner = AIFlowServerRunner(config_file=os.path.dirname(__file__) + '/master_config.yaml')
        server_runner.start(is_block=False)
        server_runner.stop()


if __name__ == '__main__':
    unittest.main()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
# 

# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import message_pb2 as message__pb2
from . import metric_service_pb2 as metric__service__pb2


class MetricServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.registerMetricMeta = channel.unary_unary(
                '/ai_flow.MetricService/registerMetricMeta',
                request_serializer=metric__service__pb2.MetricMetaRequest.SerializeToString,
                response_deserializer=metric__service__pb2.MetricMetaResponse.FromString,
                )
        self.updateMetricMeta = channel.unary_unary(
                '/ai_flow.MetricService/updateMetricMeta',
                request_serializer=metric__service__pb2.MetricMetaRequest.SerializeToString,
                response_deserializer=metric__service__pb2.MetricMetaResponse.FromString,
                )
        self.deleteMetricMeta = channel.unary_unary(
                '/ai_flow.MetricService/deleteMetricMeta',
                request_serializer=metric__service__pb2.MetricNameRequest.SerializeToString,
                response_deserializer=message__pb2.Response.FromString,
                )
        self.getMetricMeta = channel.unary_unary(
                '/ai_flow.MetricService/getMetricMeta',
                request_serializer=metric__service__pb2.MetricNameRequest.SerializeToString,
                response_deserializer=metric__service__pb2.MetricMetaResponse.FromString,
                )
        self.listDatasetMetricMetas = channel.unary_unary(
                '/ai_flow.MetricService/listDatasetMetricMetas',
                request_serializer=metric__service__pb2.ListDatasetMetricMetasRequest.SerializeToString,
                response_deserializer=metric__service__pb2.ListMetricMetasResponse.FromString,
                )
        self.listModelMetricMetas = channel.unary_unary(
                '/ai_flow.MetricService/listModelMetricMetas',
                request_serializer=metric__service__pb2.ListModelMetricMetasRequest.SerializeToString,
                response_deserializer=metric__service__pb2.ListMetricMetasResponse.FromString,
                )
        self.registerMetricSummary = channel.unary_unary(
                '/ai_flow.MetricService/registerMetricSummary',
                request_serializer=metric__service__pb2.MetricSummaryRequest.SerializeToString,
                response_deserializer=metric__service__pb2.MetricSummaryResponse.FromString,
                )
        self.updateMetricSummary = channel.unary_unary(
                '/ai_flow.MetricService/updateMetricSummary',
                request_serializer=metric__service__pb2.MetricSummaryRequest.SerializeToString,
                response_deserializer=metric__service__pb2.MetricSummaryResponse.FromString,
                )
        self.deleteMetricSummary = channel.unary_unary(
                '/ai_flow.MetricService/deleteMetricSummary',
                request_serializer=metric__service__pb2.UuidRequest.SerializeToString,
                response_deserializer=message__pb2.Response.FromString,
                )
        self.getMetricSummary = channel.unary_unary(
                '/ai_flow.MetricService/getMetricSummary',
                request_serializer=metric__service__pb2.UuidRequest.SerializeToString,
                response_deserializer=metric__service__pb2.MetricSummaryResponse.FromString,
                )
        self.listMetricSummaries = channel.unary_unary(
                '/ai_flow.MetricService/listMetricSummaries',
                request_serializer=metric__service__pb2.ListMetricSummariesRequest.SerializeToString,
                response_deserializer=metric__service__pb2.ListMetricSummariesResponse.FromString,
                )


class MetricServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def registerMetricMeta(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def updateMetricMeta(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def deleteMetricMeta(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getMetricMeta(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def listDatasetMetricMetas(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def listModelMetricMetas(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def registerMetricSummary(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def updateMetricSummary(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def deleteMetricSummary(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getMetricSummary(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def listMetricSummaries(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_MetricServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'registerMetricMeta': grpc.unary_unary_rpc_method_handler(
                    servicer.registerMetricMeta,
                    request_deserializer=metric__service__pb2.MetricMetaRequest.FromString,
                    response_serializer=metric__service__pb2.MetricMetaResponse.SerializeToString,
            ),
            'updateMetricMeta': grpc.unary_unary_rpc_method_handler(
                    servicer.updateMetricMeta,
                    request_deserializer=metric__service__pb2.MetricMetaRequest.FromString,
                    response_serializer=metric__service__pb2.MetricMetaResponse.SerializeToString,
            ),
            'deleteMetricMeta': grpc.unary_unary_rpc_method_handler(
                    servicer.deleteMetricMeta,
                    request_deserializer=metric__service__pb2.MetricNameRequest.FromString,
                    response_serializer=message__pb2.Response.SerializeToString,
            ),
            'getMetricMeta': grpc.unary_unary_rpc_method_handler(
                    servicer.getMetricMeta,
                    request_deserializer=metric__service__pb2.MetricNameRequest.FromString,
                    response_serializer=metric__service__pb2.MetricMetaResponse.SerializeToString,
            ),
            'listDatasetMetricMetas': grpc.unary_unary_rpc_method_handler(
                    servicer.listDatasetMetricMetas,
                    request_deserializer=metric__service__pb2.ListDatasetMetricMetasRequest.FromString,
                    response_serializer=metric__service__pb2.ListMetricMetasResponse.SerializeToString,
            ),
            'listModelMetricMetas': grpc.unary_unary_rpc_method_handler(
                    servicer.listModelMetricMetas,
                    request_deserializer=metric__service__pb2.ListModelMetricMetasRequest.FromString,
                    response_serializer=metric__service__pb2.ListMetricMetasResponse.SerializeToString,
            ),
            'registerMetricSummary': grpc.unary_unary_rpc_method_handler(
                    servicer.registerMetricSummary,
                    request_deserializer=metric__service__pb2.MetricSummaryRequest.FromString,
                    response_serializer=metric__service__pb2.MetricSummaryResponse.SerializeToString,
            ),
            'updateMetricSummary': grpc.unary_unary_rpc_method_handler(
                    servicer.updateMetricSummary,
                    request_deserializer=metric__service__pb2.MetricSummaryRequest.FromString,
                    response_serializer=metric__service__pb2.MetricSummaryResponse.SerializeToString,
            ),
            'deleteMetricSummary': grpc.unary_unary_rpc_method_handler(
                    servicer.deleteMetricSummary,
                    request_deserializer=metric__service__pb2.UuidRequest.FromString,
                    response_serializer=message__pb2.Response.SerializeToString,
            ),
            'getMetricSummary': grpc.unary_unary_rpc_method_handler(
                    servicer.getMetricSummary,
                    request_deserializer=metric__service__pb2.UuidRequest.FromString,
                    response_serializer=metric__service__pb2.MetricSummaryResponse.SerializeToString,
            ),
            'listMetricSummaries': grpc.unary_unary_rpc_method_handler(
                    servicer.listMetricSummaries,
                    request_deserializer=metric__service__pb2.ListMetricSummariesRequest.FromString,
                    response_serializer=metric__service__pb2.ListMetricSummariesResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'ai_flow.MetricService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class MetricService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def registerMetricMeta(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.MetricService/registerMetricMeta',
            metric__service__pb2.MetricMetaRequest.SerializeToString,
            metric__service__pb2.MetricMetaResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def updateMetricMeta(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.MetricService/updateMetricMeta',
            metric__service__pb2.MetricMetaRequest.SerializeToString,
            metric__service__pb2.MetricMetaResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def deleteMetricMeta(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.MetricService/deleteMetricMeta',
            metric__service__pb2.MetricNameRequest.SerializeToString,
            message__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getMetricMeta(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.MetricService/getMetricMeta',
            metric__service__pb2.MetricNameRequest.SerializeToString,
            metric__service__pb2.MetricMetaResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def listDatasetMetricMetas(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.MetricService/listDatasetMetricMetas',
            metric__service__pb2.ListDatasetMetricMetasRequest.SerializeToString,
            metric__service__pb2.ListMetricMetasResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def listModelMetricMetas(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.MetricService/listModelMetricMetas',
            metric__service__pb2.ListModelMetricMetasRequest.SerializeToString,
            metric__service__pb2.ListMetricMetasResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def registerMetricSummary(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.MetricService/registerMetricSummary',
            metric__service__pb2.MetricSummaryRequest.SerializeToString,
            metric__service__pb2.MetricSummaryResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def updateMetricSummary(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.MetricService/updateMetricSummary',
            metric__service__pb2.MetricSummaryRequest.SerializeToString,
            metric__service__pb2.MetricSummaryResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def deleteMetricSummary(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.MetricService/deleteMetricSummary',
            metric__service__pb2.UuidRequest.SerializeToString,
            message__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getMetricSummary(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.MetricService/getMetricSummary',
            metric__service__pb2.UuidRequest.SerializeToString,
            metric__service__pb2.MetricSummaryResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def listMetricSummaries(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.MetricService/listMetricSummaries',
            metric__service__pb2.ListMetricSummariesRequest.SerializeToString,
            metric__service__pb2.ListMetricSummariesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from ai_flow.graph.node import Node
from typing import Dict, List, Text, Optional
from ai_flow.graph.edge import Edge
from ai_flow.util.json_utils import loads


class Graph(Node):
    """
    The graph is composed of nodes(ai_flow.graph.node.Node)
    and the nodes are connected by edges(ai_flow.graph.edge.Edge).
    """

    def __init__(self) -> None:
        super().__init__()
        self.nodes: Dict[Text, Node] = {}
        self.edges: Dict[Text, List[Edge]] = {}

    def add_node(self, node: Node):
        """Add node to the graph"""
        self.nodes[node.node_id] = node

    def add_edge(self, node_id: Text, edge: Edge):
        """Add edge to the graph"""
        if node_id in self.edges:
            for e in self.edges[node_id]:
                if e == edge:
                    return
            self.edges[node_id].append(edge)
        else:
            self.edges[node_id] = []
            self.edges[node_id].append(edge)

    def is_in_graph(self, node_id: Text) -> bool:
        """Returns whether the node is in the nods of the graph"""
        return node_id in self.nodes

    def get_node_by_id(self, node_id: Text) -> Optional[Node]:
        """Return the node which node_id field equals parameter node_id"""
        if node_id in self.nodes:
            return self.nodes[node_id]
        else:
            return None

    def clear_graph(self):
        """Remove all nodes and edges in the graph."""
        self.nodes.clear()
        self.edges.clear()

    def is_empty(self) -> bool:
        """Return if the graph has nodes or edges."""
        return len(self.nodes) == 0 and len(self.edges) == 0


def load_graph(json_text: str) -> Graph:
    """Load the graph from the json string(json_text)"""
    graph: Graph = loads(json_text)
    return graph



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import os
from typing import List, Text
from ai_flow.project.project_config import ProjectConfig
from ai_flow.util.json_utils import Jsonable


class ProjectContext(Jsonable):
    """
    ProjectContext maintains the directory structure and information of an ai flow project.
    A common ai flow project would follow the directory structure as bellow.

    SimpleProject
        |- workflows # Directory for storing workflow definitions
           |- workflow_name1 # Is a directory store workflow named workflow_name1.
           |- workflow_name2 # Is a directory store workflow named workflow_name2.
           |- workflow_name3 # Is a directory store workflow named workflow_name3.
              |- workflow_name3.py # Workflow defined file.
              |- workflow_name3.yaml # Workflow configuration file.
        |- dependencies # Resource storage directory on which workflow execution depends.
            |-python # Python file resource storage directory on which workflow execution depends.
            |-jar # Jar resource storage directory on which workflow execution depends.
            |-go #  Go resource storage directory on which workflow execution depends.
        |- generated # The File storage directory generated in the process of translator generating workflow.
        |- resources # Store the resource files of the project.
        └─ project.yaml # The project configuration file.
    """

    def __init__(self) -> None:
        super().__init__()
        self.project_path: Text = None
        self.project_config: ProjectConfig = None

    @property
    def project_name(self):
        return self.project_config.get_project_name()

    def get_resources_path(self) -> Text:
        return os.path.join(self.project_path, 'resources')

    def get_generated_path(self) -> Text:
        return os.path.join(self.project_path, 'generated')

    def list_resources_paths(self) -> List:
        return os.listdir(self.get_resources_path())

    def get_dependencies_path(self) -> Text:
        return os.path.join(self.project_path, 'dependencies')

    def get_project_config_file(self) -> Text:
        return os.path.join(self.project_path, 'project.yaml')

    def get_python_dependencies_path(self) -> Text:
        return os.path.join(self.get_dependencies_path(), 'python')

    def get_jar_dependencies_path(self) -> Text:
        return os.path.join(self.get_dependencies_path(), 'jar')

    def list_jar_paths(self) -> List:
        return os.listdir(self.get_jar_dependencies_path())

    def get_go_dependencies_path(self) -> Text:
        return os.path.join(self.get_dependencies_path(), 'go')

    def get_workflows_path(self) -> Text:
        return os.path.join(self.project_path, 'workflows')

    def list_workflows(self) -> List[Text]:
        return os.listdir(self.get_workflows_path())

    def get_workflow_path(self, workflow_name) -> Text:
        return os.path.join(self.get_workflows_path(), workflow_name)

    def get_workflow_config_file(self, workflow_name) -> Text:
        return os.path.join(self.get_workflow_path(workflow_name), '{}.yaml'.format(workflow_name))

    def get_workflow_entry_file(self, workflow_name) -> Text:
        return os.path.join(self.get_workflow_path(workflow_name), '{}.py'.format(workflow_name))

    def get_workflow_entry_module(self, workflow_name) -> Text:
        return workflow_name


def build_project_context(project_path: Text) -> ProjectContext:
    """
    Load a project context for a given project path.
    :param project_path: the path of a ai flow project.
    :return: a ProjectContext object that contains the structure information of this project.
    """
    project_context = ProjectContext()
    project_path = os.path.abspath(project_path)
    project_context.project_path = project_path
    project_context.project_config = ProjectConfig()
    project_context.project_config.load_from_file(project_context.get_project_config_file())
    return project_context


__current_project_context__ = ProjectContext()

__current_project_config__ = ProjectConfig()


def init_project_config(project_config_file):
    """
    Load project configuration of the ai flow project.
    """
    __current_project_config__.load_from_file(project_config_file)


def init_project_context(project_path: Text):
    """
    Load project configuration and project context of the ai flow project.
    """
    global __current_project_context__, __current_project_config__
    project_context = build_project_context(project_path)
    __current_project_context__ = project_context
    init_project_config(project_context.get_project_config_file())


def current_project_context() -> ProjectContext:
    """
    return: The current project context(ai_flow.context.project_context.ProjectContext).
    """
    return __current_project_context__


def current_project_config() -> ProjectConfig:
    """
    :return: The current project configuration(ai_flow.project.project_config.ProjectConfig)
    """
    return __current_project_config__



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import ast
from typing import Text, Optional, Union, List
from ai_flow.common.properties import Properties
from ai_flow.endpoint.server import stringValue, int64Value
from ai_flow.meta.metric_meta import MetricMeta, MetricType, MetricSummary
from ai_flow.protobuf.message_pb2 import MetricMetaProto, MetricSummaryProto, MetricTypeProto, ReturnCode, \
    RESOURCE_DOES_NOT_EXIST, SUCCESS
from ai_flow.protobuf.metric_service_pb2 import MetricMetaResponse, MetricSummaryResponse, ListMetricSummariesResponse, \
    ListMetricMetasResponse
from ai_flow.store.db.db_model import SqlMetricMeta, SqlMetricSummary, MongoMetricMeta, MongoMetricSummary


def table_to_metric_meta(metric_meta_result) -> MetricMeta:
    properties = metric_meta_result.properties
    if properties is not None:
        properties = ast.literal_eval(properties)
    return MetricMeta(metric_name=metric_meta_result.metric_name,
                      metric_type=MetricType.value_of(metric_meta_result.metric_type),
                      metric_desc=metric_meta_result.metric_desc,
                      project_name=metric_meta_result.project_name,
                      dataset_name=metric_meta_result.dataset_name,
                      model_name=metric_meta_result.model_name,
                      job_name=metric_meta_result.job_name,
                      start_time=metric_meta_result.start_time,
                      end_time=metric_meta_result.end_time,
                      uri=metric_meta_result.uri,
                      tags=metric_meta_result.tags,
                      properties=properties)


def table_to_metric_summary(metric_summary_result) -> MetricSummary:
    return MetricSummary(uuid=metric_summary_result.uuid,
                         metric_name=metric_summary_result.metric_name,
                         metric_key=metric_summary_result.metric_key,
                         metric_value=metric_summary_result.metric_value,
                         metric_timestamp=metric_summary_result.metric_timestamp,
                         model_version=metric_summary_result.model_version,
                         job_execution_id=metric_summary_result.job_execution_id)


def metric_meta_to_table(metric_name: Text,
                         metric_type: MetricType,
                         metric_desc: Optional[Text],
                         project_name: Text,
                         dataset_name: Optional[Text],
                         model_name: Optional[Text],
                         job_name: Optional[Text],
                         start_time: int,
                         end_time: int,
                         uri: Optional[Text],
                         tags: Optional[Text],
                         properties: Properties,
                         store_type: Text = 'SqlAlchemyStore'):
    if properties is not None:
        properties = str(properties)
    if store_type == 'MongoStore':
        _class = MongoMetricMeta
    else:
        _class = SqlMetricMeta
    return _class(metric_name=metric_name,
                  metric_type=metric_type.value,
                  metric_desc=metric_desc,
                  project_name=project_name,
                  dataset_name=dataset_name,
                  model_name=model_name,
                  job_name=job_name,
                  start_time=start_time,
                  end_time=end_time,
                  uri=uri,
                  tags=tags,
                  properties=properties)


def metric_summary_to_table(metric_name: Text,
                            metric_key: Text,
                            metric_value: Text,
                            metric_timestamp: int,
                            model_version: Optional[Text],
                            job_execution_id: Optional[Text],
                            store_type: Text = 'SqlAlchemyStore'):
    if store_type == 'MongoStore':
        _class = MongoMetricSummary
    else:
        _class = SqlMetricSummary
    return _class(metric_name=metric_name,
                  metric_key=metric_key,
                  metric_value=metric_value,
                  metric_timestamp=metric_timestamp,
                  model_version=model_version,
                  job_execution_id=job_execution_id)


def metric_meta_to_proto(metric_meta: MetricMeta) -> MetricMetaProto:
    if metric_meta.metric_type == MetricType.DATASET:
        metric_type = MetricTypeProto.DATASET
    else:
        metric_type = MetricTypeProto.MODEL
    return MetricMetaProto(metric_name=stringValue(metric_meta.metric_name),
                           metric_type=metric_type,
                           metric_desc=stringValue(metric_meta.metric_desc),
                           project_name=stringValue(metric_meta.project_name),
                           dataset_name=stringValue(metric_meta.dataset_name),
                           model_name=stringValue(metric_meta.model_name),
                           job_name=stringValue(metric_meta.job_name),
                           start_time=int64Value(metric_meta.start_time),
                           end_time=int64Value(metric_meta.end_time),
                           uri=stringValue(metric_meta.uri),
                           tags=stringValue(metric_meta.tags),
                           properties=metric_meta.properties)


def metric_summary_to_proto(metric_summary: MetricSummary) -> MetricSummaryProto:
    return MetricSummaryProto(uuid=metric_summary.uuid,
                              metric_name=stringValue(metric_summary.metric_name),
                              metric_key=stringValue(metric_summary.metric_key),
                              metric_value=stringValue(metric_summary.metric_value),
                              metric_timestamp=int64Value(metric_summary.metric_timestamp),
                              model_version=stringValue(metric_summary.model_version),
                              job_execution_id=stringValue(metric_summary.job_execution_id))


def proto_to_metric_meta(metric_meta_proto: MetricMetaProto) -> MetricMeta:
    if MetricTypeProto.DATASET == metric_meta_proto.metric_type:
        metric_type = MetricType.DATASET
    else:
        metric_type = MetricType.MODEL
    return MetricMeta(
        metric_name=metric_meta_proto.metric_name.value if metric_meta_proto.HasField('metric_name') else None,
        metric_type=metric_type,
        metric_desc=metric_meta_proto.metric_desc.value
        if metric_meta_proto.HasField('metric_desc') else None,
        project_name=metric_meta_proto.project_name.value,
        dataset_name=metric_meta_proto.dataset_name.value if metric_meta_proto.HasField(
            'dataset_name') else None,
        model_name=metric_meta_proto.model_name.value if metric_meta_proto.HasField(
            'model_name') else None,
        job_name=metric_meta_proto.job_name.value if metric_meta_proto.HasField('job_name') else None,
        start_time=metric_meta_proto.start_time.value if metric_meta_proto.HasField(
            'start_time') else None,
        end_time=metric_meta_proto.end_time.value if metric_meta_proto.HasField(
            'end_time') else None,
        uri=metric_meta_proto.uri.value if metric_meta_proto.HasField('uri') else None,
        tags=metric_meta_proto.tags.value if metric_meta_proto.HasField('tags') else None,
        properties=None if metric_meta_proto.properties == {} else metric_meta_proto.properties,
    )


def proto_to_metric_summary(metric_summary_proto: MetricSummaryProto) -> MetricSummary:
    return MetricSummary(uuid=metric_summary_proto.uuid,
                         metric_name=metric_summary_proto.metric_name.value if metric_summary_proto.HasField(
                             'metric_name') else None,
                         metric_key=metric_summary_proto.metric_key.value if metric_summary_proto.HasField(
                             'metric_key') else None,
                         metric_value=metric_summary_proto.metric_value.value if metric_summary_proto.HasField(
                             'metric_value') else None,
                         metric_timestamp=metric_summary_proto.metric_timestamp.value if metric_summary_proto.HasField(
                             'metric_timestamp') else None,
                         model_version=metric_summary_proto.model_version.value if metric_summary_proto.HasField(
                             'model_version') else None,
                         job_execution_id=metric_summary_proto.job_execution_id.value if metric_summary_proto.HasField(
                             'job_execution_id') else None
                         )


def _warp_metric_meta_response(metric_meta: Optional[MetricMeta]) -> MetricMetaResponse:
    if metric_meta is not None:
        return MetricMetaResponse(return_code=0, return_msg=ReturnCode.Name(SUCCESS).lower(),
                                  metric_meta=metric_meta_to_proto(metric_meta))
    else:
        return MetricMetaResponse(return_code=1,
                                  return_msg=ReturnCode.Name(RESOURCE_DOES_NOT_EXIST).lower(),
                                  metric_meta=None)


def _warp_list_metric_metas_response(metric_metas: Union[None, MetricMeta, List[MetricMeta]]) -> MetricMetaResponse:
    if metric_metas is not None:
        if isinstance(metric_metas, MetricMeta):
            return ListMetricMetasResponse(return_code=0, return_msg=ReturnCode.Name(SUCCESS).lower(),
                                           metric_metas=[metric_meta_to_proto(metric_metas)])
        else:
            res = []
            for metric_meta in metric_metas:
                res.append(metric_meta_to_proto(metric_meta))
            return ListMetricMetasResponse(return_code=0, return_msg=ReturnCode.Name(SUCCESS).lower(),
                                           metric_metas=res)
    else:
        return ListMetricMetasResponse(return_code=1,
                                       return_msg=ReturnCode.Name(RESOURCE_DOES_NOT_EXIST).lower(),
                                       metric_metas=None)


def _warp_metric_summary_response(metric_summary: Optional[MetricSummary]) -> MetricSummaryResponse:
    if metric_summary is not None:
        return MetricSummaryResponse(return_code=0, return_msg=ReturnCode.Name(SUCCESS).lower(),
                                     metric_summary=metric_summary_to_proto(metric_summary))
    else:
        return MetricSummaryResponse(return_code=1,
                                     return_msg=ReturnCode.Name(RESOURCE_DOES_NOT_EXIST).lower(),
                                     metric_summary=None)


def _warp_list_metric_summaries_response(
        metric_summaries: Union[None, MetricSummary, List[MetricSummary]]) -> ListMetricSummariesResponse:
    if metric_summaries is not None:
        if isinstance(metric_summaries, MetricSummary):
            return ListMetricSummariesResponse(return_code=0, return_msg=ReturnCode.Name(SUCCESS).lower(),
                                               metric_summaries=[metric_summary_to_proto(metric_summaries)])
        else:
            res = []
            for metric_summary in metric_summaries:
                res.append(metric_summary_to_proto(metric_summary))
            return ListMetricSummariesResponse(return_code=0, return_msg=ReturnCode.Name(SUCCESS).lower(),
                                               metric_summaries=res)
    else:
        return ListMetricSummariesResponse(return_code=1,
                                           return_msg=ReturnCode.Name(RESOURCE_DOES_NOT_EXIST).lower(),
                                           metric_summaries=None)
